# -*- coding: utf-8 -*-

# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Evaluation Orchestration Library."""

import collections
from concurrent import futures
import copy
import time
from typing import Any, Callable, Dict, List, Optional, Tuple, TYPE_CHECKING, Union

from google.cloud.aiplatform import base
from google.cloud.aiplatform_v1beta1.types import (
    content as gapic_content_types,
)
from vertexai import generative_models
from vertexai.preview import reasoning_engines
from vertexai.preview.evaluation import _base as evaluation_base
from vertexai.preview.evaluation import _pre_eval_utils
from vertexai.preview.evaluation import constants
from vertexai.preview.evaluation import (
    prompt_template as prompt_template_base,
)
from vertexai.preview.evaluation import utils
from vertexai.preview.evaluation.metrics import (
    _base as metrics_base,
)
from vertexai.preview.evaluation.metrics import (
    _instance_evaluation,
)
from vertexai.preview.evaluation.metrics import (
    custom_output_config,
)
from vertexai.preview.evaluation.metrics import (
    metric_prompt_template_examples,
)
from vertexai.preview.evaluation.metrics import pairwise_metric
from vertexai.preview.evaluation.metrics import pointwise_metric
from vertexai.preview.evaluation.metrics import (
    rubric_based_metric,
)


if TYPE_CHECKING:
    import pandas as pd

try:
    from tqdm import tqdm
except ImportError:
    raise ImportError(
        'tqdm is not installed. Please install the SDK using "pip install'
        ' google-cloud-aiplatform[evaluation]"'
    )

_LOGGER = base.Logger(__name__)
_SUCCESSFUL_FINISH_REASONS = [
    gapic_content_types.Candidate.FinishReason.STOP,
    gapic_content_types.Candidate.FinishReason.MAX_TOKENS,
    # Many responses have this finish reason
    gapic_content_types.Candidate.FinishReason.FINISH_REASON_UNSPECIFIED,
]

_RunnableType = Union[reasoning_engines.Queryable, Callable[[str], Dict[str, str]]]
_ModelType = Union[generative_models.GenerativeModel, Callable[[str], str]]


def _validate_metrics(metrics: List[Union[str, metrics_base._Metric]]) -> None:
    """Validates the metrics list.

    Args:
      metrics: The list of metric names, or Metric instances to
        evaluate.

    Raises:
      ValueError: If metric is empty or if multiple metrics of the
        same metric name are found.
    """
    if not metrics:
        raise ValueError("Metrics cannot be empty.")

    seen_strings = set()
    seen_metric_names = set()
    rubric_based_metric_present = False

    for metric in metrics:
        if isinstance(metric, str):
            if metric in seen_strings:
                raise ValueError(f"Duplicate string metric name found: '{metric}'")
            seen_strings.add(metric)
        elif isinstance(
            metric, metrics_base._Metric
        ):  # pylint: disable=protected-access
            if metric.metric_name in seen_metric_names:
                raise ValueError(
                    "Duplicate Metric instances of the same metric name found: "
                    f"'{metric.metric_name}'"
                )
            if isinstance(metric, rubric_based_metric.RubricBasedMetric):
                if rubric_based_metric_present:
                    raise ValueError("Multiple rubric based metrics are not supported.")
                rubric_based_metric_present = True
            seen_metric_names.add(metric.metric_name)


def _validate_metric_column_map(
    evaluation_run_config: evaluation_base.EvaluationRunConfig,
):
    """Validates the column map for metric prompt template usage."""
    for metric in evaluation_run_config.metrics:
        if isinstance(
            metric, metrics_base._ModelBasedMetric  # pylint: disable=protected-access
        ):
            for variable in prompt_template_base.PromptTemplate(
                metric.metric_prompt_template
            ).variables:
                if (
                    evaluation_run_config.metric_column_mapping.get(variable, "")
                    not in evaluation_run_config.dataset.columns
                ):
                    raise ValueError(
                        f"Cannot find the `{variable}` column in the evaluation"
                        " dataset to fill the metric prompt template for"
                        f" `{str(metric)}` metric. Please check if the column is"
                        " present in the evaluation dataset, or provide a"
                        " key-value pair in `metric_column_mapping` parameter"
                        " of `EvalTask` to map it to a different column name."
                        " The evaluation dataset columns are"
                        f" {list(evaluation_run_config.dataset.columns)}."
                    )


def _validate_autorater_config(
    evaluation_run_config: evaluation_base.EvaluationRunConfig,
    pairwise_metric_exists: bool = False,
):
    """Validates the fields in the autorater config.

    Args:
      evaluation_run_config: The evaluation run config.
      pairwise_metric_exists: Whether pairwise metrics are used.

    Raises:
      ValueError: If the sampling count is not in the range [1, 32], or flipping
      is enabled for pairwise metrics but the required columns are not found in
      the dataset.
    """
    if not evaluation_run_config.autorater_config:
        return
    if evaluation_run_config.autorater_config.sampling_count and (
        evaluation_run_config.autorater_config.sampling_count < 1
        or evaluation_run_config.autorater_config.sampling_count > 32
    ):
        raise ValueError(
            "autorater_config.sampling_count must be in the range [1, 32]."
        )

    if pairwise_metric_exists and evaluation_run_config.autorater_config.flip_enabled:
        for column in [
            constants.Dataset.MODEL_RESPONSE_COLUMN,
            constants.Dataset.BASELINE_MODEL_RESPONSE_COLUMN,
        ]:
            field_name = evaluation_run_config.metric_column_mapping.get(column, "")
            if (
                not field_name
                or field_name not in evaluation_run_config.dataset.columns
            ):
                raise ValueError(
                    f"Cannot find the `{column}` column in"
                    " the evaluation dataset to fill the metric prompt template."
                    " This is required for Pairwise metrics when performing"
                    " autorater config flipping. Please check if the column is"
                    " present in the evaluation dataset, or provide a key-value"
                    " pair in `metric_column_mapping` parameter of `EvalTask` to"
                    " map it to a different column name. The evaluation dataset"
                    f" columns are {list(evaluation_run_config.dataset.columns)}."
                )


def _validate_column_provided(
    evaluation_run_config: evaluation_base.EvaluationRunConfig,
    column_name: str,
) -> None:
    """Validates the required column exist in the dataset."""
    if column_name not in evaluation_run_config.metric_column_mapping:
        evaluation_run_config.metric_column_mapping[column_name] = column_name
    evaluation_run_config.validate_dataset_column(column_name)


def _validate_response_column_required(
    evaluation_run_config: evaluation_base.EvaluationRunConfig,
) -> None:
    """Validates the response column exists in the dataset."""
    for metric in evaluation_run_config.metrics:
        if metric in constants.Metric.AUTOMATIC_METRIC_LIST:
            _validate_column_provided(
                evaluation_run_config,
                constants.Dataset.MODEL_RESPONSE_COLUMN,
            )


def _validate_reference_column_required(
    evaluation_run_config: evaluation_base.EvaluationRunConfig,
) -> None:
    """Validates the reference column exists in the dataset."""
    if set(evaluation_run_config.metrics).intersection(
        set(constants.Metric.AUTOMATIC_METRIC_LIST)
    ):
        _validate_column_provided(
            evaluation_run_config,
            constants.Dataset.REFERENCE_COLUMN,
        )


def _validate_dataset(
    evaluation_run_config: evaluation_base.EvaluationRunConfig,
) -> None:
    """Validates the required columns exists in the dataset."""
    _validate_response_column_required(evaluation_run_config)
    _validate_reference_column_required(evaluation_run_config)


def _compute_custom_metrics(
    row_dict: Dict[str, Any],
    custom_metrics: List[metrics_base.CustomMetric],
    pbar: tqdm,
    executor: futures.ThreadPoolExecutor,
) -> Dict[str, Any]:
    """Computes custom metrics for a row.

    Args:
        row_dict: A dictionary of an instance in the eval dataset.
        custom_metrics: A list of CustomMetrics.
        pbar: A tqdm progress bar.
        executor: A thread pool executor.

    Returns:
        A dictionary of an instance containing custom metric results.

    Raises:
        KeyError: If the custom metric function does not return a valid output.
    """
    futures_by_metric = collections.defaultdict(list)
    for custom_metric in custom_metrics:
        future = executor.submit(custom_metric.metric_function, row_dict)
        future.add_done_callback(lambda _: pbar.update(1))
        futures_by_metric[custom_metric].append(future)

    for custom_metric, futures_list in futures_by_metric.items():
        for future in futures_list:
            metric_output = future.result()
            try:
                row_dict[f"{custom_metric.name}/{constants.MetricResult.SCORE_KEY}"] = (
                    metric_output[custom_metric.name]
                )
            except KeyError:
                raise KeyError(
                    f"Custom metric score `{custom_metric.name}` not found in"
                    f" the metric output {metric_output}. Please make sure the"
                    " custom metric function is valid, and the output"
                    f" dictionary uses `{custom_metric.name}` as the key for"
                    " metric score."
                )
            # Include additional metric results like explanation.
            for key, value in metric_output.items():
                if key != custom_metric.name:
                    row_dict[f"{custom_metric.name}/{key}"] = value
    return row_dict


def _separate_custom_metrics(
    metrics: List[Union[str, metrics_base._Metric]],
) -> Tuple[
    List[Union[str, metrics_base._Metric]],
    List[metrics_base.CustomMetric],
]:
    """Separates the metrics list into API and custom metrics."""
    custom_metrics = []
    api_metrics = []
    for metric in metrics:
        if isinstance(metric, metrics_base.CustomMetric):
            custom_metrics.append(metric)
        else:
            api_metrics.append(metric)
    return api_metrics, custom_metrics


def _aggregate_summary_metrics(
    evaluation_run_config: evaluation_base.EvaluationRunConfig,
    metrics_table: "pd.DataFrame",
) -> Dict[str, Any]:
    """Computes summary metrics.

    Args:
        evaluation_run_config: Evaluation Run Configurations.
        metrics_table: A dataframe containing per-instance metrics results.

    Returns:
        A dictionary containing summary metrics results and statistics.
    """
    summary_metrics = {}
    summary_metrics[constants.MetricResult.ROW_COUNT_KEY] = metrics_table.shape[0]

    for metric in evaluation_run_config.metrics:
        try:
            if isinstance(metric, pairwise_metric.PairwiseMetric):
                pairwise_choice_col_name = (
                    f"{metric.metric_name}/{constants.MetricResult.PAIRWISE_CHOICE_KEY}"
                )
                if pairwise_choice_col_name in metrics_table:
                    candidate_model_win_rate_choices = [
                        "CANDIDATE",
                        utils.RATING_TO_VERDICT["B>>A"],
                        utils.RATING_TO_VERDICT["A<<B"],
                        utils.RATING_TO_VERDICT["B>A"],
                        utils.RATING_TO_VERDICT["A<B"],
                    ]
                    baseline_model_win_rate_choices = [
                        "BASELINE",
                        utils.RATING_TO_VERDICT["A>>B"],
                        utils.RATING_TO_VERDICT["B<<A"],
                        utils.RATING_TO_VERDICT["B<A"],
                        utils.RATING_TO_VERDICT["A>B"],
                    ]
                    summary_metrics[
                        f"{metric.metric_name}/candidate_model_win_rate"
                    ] = (
                        metrics_table[pairwise_choice_col_name].isin(
                            candidate_model_win_rate_choices
                        )
                    ).mean()
                    summary_metrics[f"{metric.metric_name}/baseline_model_win_rate"] = (
                        metrics_table[pairwise_choice_col_name].isin(
                            baseline_model_win_rate_choices
                        )
                    ).mean()
            else:
                score_col_name = f"{str(metric)}/{constants.MetricResult.SCORE_KEY}"
                if score_col_name in metrics_table:
                    summary_metrics[f"{str(metric)}/mean"] = metrics_table.loc[
                        :, score_col_name
                    ].mean()
                    summary_metrics[f"{str(metric)}/std"] = metrics_table.loc[
                        :, score_col_name
                    ].std()
        except (ValueError, KeyError) as e:
            _LOGGER.warning(
                f"Failed to compute metric statistics for `{metric}` metric."
                f"{type(e).__name__}: {e}"
            )
            continue

    metrics_table_columns = metrics_table.columns.tolist()
    for default_metric in constants.Metric.DEFAULT_METRIC_LIST:
        if default_metric in metrics_table_columns:
            summary_metrics[f"{str(default_metric)}/mean"] = metrics_table.loc[
                :, f"{str(default_metric)}"
            ].mean()
            summary_metrics[f"{str(default_metric)}/std"] = metrics_table.loc[
                :, f"{str(default_metric)}"
            ].std()
    return summary_metrics


def _run_model_inference(
    model: _ModelType,
    evaluation_run_config: evaluation_base.EvaluationRunConfig,
    response_column_name: str = constants.Dataset.MODEL_RESPONSE_COLUMN,
) -> None:
    """Runs model inference on dataset for evaluation.

    Args:
      model: The model or baseline model or a custom model function to
        generate responses to evaluate.
      evaluation_run_config: Evaluation Run Configurations.
      response_column_name: Column name key in metric_column_mapping. Value is
        constants.Dataset.MODEL_RESPONSE_COLUMN or
        constants.Dataset.BASELINE_MODEL_RESPONSE_COLUMN.

    Raises:
        ValueError: If the model or baseline model is not supported.
    """
    is_baseline_model = (
        response_column_name == constants.Dataset.BASELINE_MODEL_RESPONSE_COLUMN
    )
    if response_column_name not in evaluation_run_config.metric_column_mapping:
        if constants.Dataset.PROMPT_COLUMN in evaluation_run_config.dataset.columns:
            t1 = time.perf_counter()
            if isinstance(model, generative_models.GenerativeModel):
                responses = _pre_eval_utils._generate_responses_from_gemini_model(
                    model, evaluation_run_config.dataset
                )
                _pre_eval_utils.populate_eval_dataset_with_model_responses(
                    responses,
                    evaluation_run_config,
                    is_baseline_model,
                )
            elif callable(model):
                responses = _pre_eval_utils._generate_response_from_custom_model_fn(
                    model, evaluation_run_config.dataset
                )
                _pre_eval_utils.populate_eval_dataset_with_model_responses(
                    responses,
                    evaluation_run_config,
                    is_baseline_model,
                )
            else:
                raise ValueError(
                    f"Unsupported model or baseline model type: {type(model)}"
                )
            t2 = time.perf_counter()
            _LOGGER.info(f"Multithreaded Batch Inference took: {t2 - t1} seconds.")
            evaluation_run_config.metric_column_mapping[response_column_name] = (
                response_column_name
            )
        else:
            raise ValueError(
                "Missing required input `prompt` column to start model inference."
                " Please provide a `prompt_template` parameter in"
                " `EvalTask.evaluate()` function if you want to assemble a"
                " `prompt` column with variables from the dataset, or provide a"
                " `prompt` column in dataset to directly use as input to"
                " the model. Mappings in `metric_column_mapping` do not"
                " apply for model inference and are used for evaluation only."
            )
    else:
        raise ValueError(
            "The `model` parameter or `baseline_model` in pairwise metric is"
            " specified, but the evaluation `dataset` contains model response"
            " column or baseline model response column"
            f" `{evaluation_run_config.metric_column_mapping[response_column_name]}`"
            " to perform bring-your-own-response(BYOR) evaluation. If you would"
            " like to perform evaluation using the dataset with the"
            " existing model response column or baseline model response column"
            f" `{evaluation_run_config.metric_column_mapping[response_column_name]}`,"
            " please remove `model` parameter in `EvalTask.evaluate()`"
            " function or `baseline_model` in `PairwiseMetric`."
        )


def _generate_response_from_custom_agent_fn(
    runnable: Callable[[str], Dict[str, str]], prompt: str
):
    """Processes a single row from the DataFrame for a custom agent."""
    t1 = time.perf_counter()
    try:
        response_dict = runnable(prompt)
        latency = time.perf_counter() - t1
        failure = 0
    except Exception as e:  # pylint: disable=broad-exception-caught
        latency = None
        response_dict = {
            "input": prompt,
            "output": f"Error occurred during inference: {e}",
            "predicted_trajectory": None,
        }
        failure = 1
    return response_dict, latency, failure


def _generate_response_from_queryable(
    runnable: reasoning_engines.Queryable, prompt: str
):
    """Processes a single row from the DataFrame for a Queryable."""
    t1 = time.perf_counter()
    try:
        response_dict = runnable.query(input=prompt)
        output_dict = {
            "input": prompt,
            "output": response_dict["output"],
        }
        if "intermediate_steps" in response_dict:
            predicted_trajectory = utils.parse_intermediate_steps(
                response_dict["intermediate_steps"]
            )
            output_dict["predicted_trajectory"] = predicted_trajectory
        latency = time.perf_counter() - t1
        failure = 0

    except Exception as e:  # pylint: disable=broad-exception-caught
        latency = None
        output_dict = {
            "input": prompt,
            "output": f"Error occurred during inference: {e}",
            "predicted_trajectory": None,
        }
        failure = 1
    return output_dict, latency, failure


def _run_runnable_inference(
    runnable: _RunnableType,
    evaluation_run_config: evaluation_base.EvaluationRunConfig,
    response_column_name: str = constants.Dataset.MODEL_RESPONSE_COLUMN,
    predicted_trajectory_column_name: str = constants.Dataset.PREDICTED_TRAJECTORY_COLUMN,
):
    """Runs inference with a runnable agent and adds the inference final response and predicted trajectory to the response column."""
    if constants.Dataset.PROMPT_COLUMN in evaluation_run_config.dataset.columns:
        responses = []
        trajectory_list = []
        latency_list = []
        failure_list = []
        tasks = []
        if isinstance(runnable, reasoning_engines.Queryable):
            with tqdm(total=len(evaluation_run_config.dataset)) as pbar:
                with futures.ThreadPoolExecutor(
                    max_workers=constants.MAX_WORKERS
                ) as executor:
                    for _, row in evaluation_run_config.dataset.iterrows():
                        task = executor.submit(
                            _generate_response_from_queryable,
                            runnable,
                            row[constants.Dataset.PROMPT_COLUMN],
                        )
                        task.add_done_callback(lambda _: pbar.update(1))
                        tasks.append(task)
                    for task in tasks:
                        response_dict, latency, failure = task.result()
                        responses.append(response_dict["output"])
                        latency_list.append(latency)
                        failure_list.append(failure)
                        if (
                            response_column_name
                            in evaluation_run_config.metric_column_mapping
                        ):
                            _LOGGER.warning(
                                "When a `runnable` is provided, response is"
                                " generated dynamically by the runnable, so "
                                "the pre-existing `response` column provided "
                                "in the evaluation dataset is not used."
                            )
                        if "predicted_trajectory" in response_dict:
                            trajectory_list.append(
                                response_dict["predicted_trajectory"]
                            )
                            if (
                                predicted_trajectory_column_name
                                in evaluation_run_config.metric_column_mapping
                            ):
                                _LOGGER.warning(
                                    "When a `runnable` is provided, trajectory is"
                                    "generated dynamically by the runnable, so "
                                    "the pre-existing `response` column provided "
                                    "in the evaluation dataset is not used."
                                )
        elif callable(runnable):
            with tqdm(total=len(evaluation_run_config.dataset)) as pbar:
                with futures.ThreadPoolExecutor(
                    max_workers=constants.MAX_WORKERS
                ) as executor:
                    for _, row in evaluation_run_config.dataset.iterrows():
                        task = executor.submit(
                            _generate_response_from_custom_agent_fn,
                            runnable,
                            row[constants.Dataset.PROMPT_COLUMN],
                        )
                        task.add_done_callback(lambda _: pbar.update(1))
                        tasks.append(task)
                    for task in tasks:
                        response_dict, latency, failure = task.result()
                        responses.append(response_dict.get("response", ""))
                        latency_list.append(latency)
                        failure_list.append(failure)
                        if "predicted_trajectory" in response_dict:
                            predicted_trajectory = response_dict["predicted_trajectory"]
                            trajectory_list.append(predicted_trajectory)
                            if (
                                response_column_name
                                in evaluation_run_config.metric_column_mapping
                            ):
                                _LOGGER.warning(
                                    "When a `runnable` is provided, response is"
                                    " generated dynamically by the runnable, so "
                                    "the pre-existing `response` column provided "
                                    "in the evaluation dataset is not used."
                                )
                            if (
                                predicted_trajectory_column_name
                                in evaluation_run_config.metric_column_mapping
                            ):
                                _LOGGER.warning(
                                    "When a `runnable` is provided, trajectory"
                                    " is generated dynamically by the runnable,"
                                    " so the pre-existing `predicted_trajectory`"
                                    " column provided in the evaluation dataset"
                                    " is not used."
                                )
        else:
            raise ValueError(f"Unsupported runnable type: {type(runnable)}")

        evaluation_run_config.dataset = evaluation_run_config.dataset.assign(
            response=responses
        )
        evaluation_run_config.dataset = evaluation_run_config.dataset.assign(
            latency_in_seconds=latency_list
        )
        evaluation_run_config.dataset = evaluation_run_config.dataset.assign(
            failure=failure_list
        )
        evaluation_run_config.metric_column_mapping[response_column_name] = (
            response_column_name
        )
        if trajectory_list:
            evaluation_run_config.dataset = evaluation_run_config.dataset.assign(
                predicted_trajectory=trajectory_list
            )
            evaluation_run_config.metric_column_mapping[
                predicted_trajectory_column_name
            ] = predicted_trajectory_column_name

        _LOGGER.info(
            f"All {evaluation_run_config.dataset.shape[0]} responses are successfully"
            " generated from the runnable."
        )
    else:
        raise ValueError(
            "Missing required input `prompt` column to start inference."
            " Please provide a `prompt_template` parameter in"
            " `EvalTask.evaluate()` function if you want to assemble a"
            " `prompt` column with variables from the dataset, or provide a"
            " `prompt` column in dataset to directly use as input to"
            " the runnable. Mappings in `metric_column_mapping` do not"
            " apply for runnable inference and are used for evaluation only."
        )


def _assemble_prompt_for_dataset(
    evaluation_run_config: evaluation_base.EvaluationRunConfig,
    prompt_template: Union[prompt_template_base.PromptTemplate, str],
) -> None:
    """Assembles a prompt column in metrics_table from variable columns.

    Args:
        evaluation_run_config: Evaluation Run Configurations.
        prompt_template:  A `PromptTemplate` object or a prompt template string
          with variables that can be assembled from the evaluation dataset. The
          variables can be represented in curly braces `{variable}`, and
          must be included in the dataset columns if specified. The variable
          names cannot contain spaces.

    Returns:
        The assembled prompt template string to send to the model.

    Raises:
        ValueError: If any variable names do not exist in the dataset columns
        or the prompt template is invalid.
    """
    if not prompt_template:
        raise ValueError("Prompt template cannot be an empty string.")

    _LOGGER.info(
        "Assembling prompts from the `prompt_template`. The `prompt` column in"
        " the `EvalResult.metrics_table` has the assembled prompts used for model"
        " response generation."
    )

    try:
        evaluation_run_config.dataset[constants.Dataset.PROMPT_COLUMN] = (
            evaluation_run_config.dataset.apply(
                lambda row: _pre_eval_utils._assemble_prompt(row, prompt_template),
                axis=1,
            )
        )
        if (
            constants.Dataset.PROMPT_COLUMN
            in evaluation_run_config.metric_column_mapping
            and evaluation_run_config.metric_column_mapping[
                constants.Dataset.PROMPT_COLUMN
            ]
            != constants.Dataset.PROMPT_COLUMN
        ):
            _LOGGER.warning(
                "The `prompt` column mapping provided in"
                " `metric_column_mapping` parameter is overwritten by the"
                " assembled `prompt` column because the `prompt_template`"
                " parameter is provided. Please verify that you want to use"
                " the assembled `prompt` column for evaluation."
            )
        evaluation_run_config.metric_column_mapping[constants.Dataset.PROMPT_COLUMN] = (
            constants.Dataset.PROMPT_COLUMN
        )
    except Exception as e:
        raise ValueError(
            f"Failed to assemble prompt template: {e}. Please make sure all"
            " variables in `prompt_template` are present in the evaluation"
            f" dataset columns: `{list(evaluation_run_config.dataset.columns)}`."
        ) from e


def _set_metric_table(
    metric_name: str,
    metric_results: Any,
    metrics_table: "pd.DataFrame",
    metric_result_key: str,
):
    """Parses value from metric results to metrics_table."""
    if (
        metric_result_key == constants.MetricResult.SCORE_KEY
        or metric_result_key == constants.MetricResult.IS_UNSAFE_PROBABILITY_KEY
    ):
        metric_result_items = [
            result.get(metric_result_key) if isinstance(result, dict) else None
            for result in metric_results
        ]
    else:
        metric_result_items = [
            result.get(metric_result_key) if isinstance(result, dict) else "Error"
            for result in metric_results
        ]
    metrics_table[f"{metric_name}/{metric_result_key}"] = metric_result_items


def _parse_metric_results_to_dataframe(
    instance_df: "pd.DataFrame", results: Dict[Union[str, metrics_base._Metric], Any]
) -> Dict[str, Any]:
    """Parses metric results to a pandas dataframe.

    Args:
        instance_df: A dataframe containing per-instance metrics results.
        results: A dictionary containing metric results.

    Returns:
        A dataframe containing per-instance metrics results. Each metric result
        can contain metric score, explanation, and confidence. If
        custom_output_config is specified, raw output or parsed output will
        be included instead.
    """
    try:
        import pandas as pd
    except ImportError:
        raise ImportError(
            'Pandas is not installed. Please install the SDK using "pip install'
            ' google-cloud-aiplatform[evaluation]"'
        )

    metrics_table = pd.DataFrame(dict(zip(instance_df.columns, instance_df.values.T)))
    for metric, metric_results in results.items():
        if isinstance(metric, pointwise_metric.PointwiseMetric):
            if getattr(metric, "custom_output_config", None) and getattr(
                metric.custom_output_config, "return_raw_output", False
            ):
                if getattr(metric, "custom_output_config", None) and getattr(
                    metric.custom_output_config, "parsing_fn", None
                ):
                    for key in metric_results[0].keys():
                        _set_metric_table(
                            metric.metric_name,
                            metric_results,
                            metrics_table,
                            key,
                        )
                else:
                    _set_metric_table(
                        metric.metric_name,
                        metric_results,
                        metrics_table,
                        constants.MetricResult.RAW_OUTPUT_KEY,
                    )
            else:
                _set_metric_table(
                    metric.metric_name,
                    metric_results,
                    metrics_table,
                    constants.MetricResult.EXPLANATION_KEY,
                )
                _set_metric_table(
                    metric.metric_name,
                    metric_results,
                    metrics_table,
                    constants.MetricResult.SCORE_KEY,
                )
        elif isinstance(metric, pairwise_metric.PairwiseMetric):
            if getattr(metric, "custom_output_config", None) and getattr(
                metric.custom_output_config, "return_raw_output", False
            ):
                if getattr(metric, "custom_output_config", None) and getattr(
                    metric.custom_output_config, "parsing_fn", None
                ):
                    for key in metric_results[0].keys():
                        _set_metric_table(
                            metric.metric_name,
                            metric_results,
                            metrics_table,
                            key,
                        )
                else:
                    _set_metric_table(
                        metric.metric_name,
                        metric_results,
                        metrics_table,
                        constants.MetricResult.RAW_OUTPUT_KEY,
                    )
            else:
                _set_metric_table(
                    metric.metric_name,
                    metric_results,
                    metrics_table,
                    constants.MetricResult.EXPLANATION_KEY,
                )
                _set_metric_table(
                    metric.metric_name,
                    metric_results,
                    metrics_table,
                    constants.MetricResult.PAIRWISE_CHOICE_KEY,
                )
        elif (
            str(metric)
            in constants.Metric.AUTOMATIC_METRIC_LIST
            + constants.Metric.TRAJECTORY_METRIC_LIST
        ):
            _set_metric_table(
                str(metric),
                metric_results,
                metrics_table,
                constants.MetricResult.SCORE_KEY,
            )
        elif metric == constants.Metric.RUBRIC_BASED_INSTRUCTION_FOLLOWING:
            _set_metric_table(
                constants.Metric.RUBRIC_BASED_INSTRUCTION_FOLLOWING,
                metric_results,
                metrics_table,
                constants.MetricResult.RUBRIC_LEVEL_INSTRUCTION_FOLLOWING_KEY,
            )
            _set_metric_table(
                constants.Metric.RUBRIC_BASED_INSTRUCTION_FOLLOWING,
                metric_results,
                metrics_table,
                constants.MetricResult.SCORE_KEY,
            )
        else:
            _LOGGER.warning(
                f"Metric name: {str(metric)} is not supported when parsing"
                " metric results."
            )

    return metrics_table


def _compute_metrics(
    evaluation_run_config: evaluation_base.EvaluationRunConfig,
) -> Tuple[Dict[str, Any], "pd.DataFrame"]:
    """Computes the metrics for the dataset.

    Args:
      evaluation_run_config: Evaluation Run Configurations.

    Returns:
      The evaluation results for the input metrics.

    Raises:
      RuntimeError: The number of responses does not match the number of metrics.
    """
    try:
        import pandas as pd
    except ImportError:
        raise ImportError(
            'Pandas is not installed. Please install the SDK using "pip install'
            ' google-cloud-aiplatform[evaluation]"'
        )

    api_metrics, custom_metrics = _separate_custom_metrics(
        evaluation_run_config.metrics
    )
    row_count = len(evaluation_run_config.dataset)
    api_request_count = len(api_metrics) * row_count
    custom_metric_request_count = len(custom_metrics) * row_count
    total_request_count = api_request_count + custom_metric_request_count

    _LOGGER.info(
        f"Computing metrics with a total of {total_request_count} Vertex Gen AI"
        " Evaluation Service API requests."
    )

    instance_list = []
    futures_by_metric = collections.defaultdict(list)
    rate_limiter = utils.RateLimiter(evaluation_run_config.evaluation_service_qps)
    results_dict = collections.defaultdict(
        lambda: [None] * len(evaluation_run_config.dataset)
    )
    error_list = []
    with tqdm(total=total_request_count) as pbar:
        with futures.ThreadPoolExecutor(max_workers=constants.MAX_WORKERS) as executor:
            for idx, row in evaluation_run_config.dataset.iterrows():
                row_dict = _compute_custom_metrics(
                    row.to_dict(), custom_metrics, pbar, executor
                )
                instance_list.append(row_dict)
                for metric in api_metrics:
                    try:
                        future = executor.submit(
                            _instance_evaluation.evaluate_instances,
                            client=evaluation_run_config.client,
                            request=_instance_evaluation.build_request(
                                metric=metric,
                                row_dict=row_dict,
                                evaluation_run_config=evaluation_run_config,
                            ),
                            rate_limiter=rate_limiter,
                            retry_timeout=evaluation_run_config.retry_timeout,
                        )
                        future.add_done_callback(lambda _: pbar.update(1))
                        futures_by_metric[metric].append((future, idx))
                    except ValueError as e:
                        results_dict[metric][idx] = "Error"
                        error_list.append((metric, idx, f"Error: {e}"))

        # Retrieve results from all futures and handle errors.
        for metric, futures_list in futures_by_metric.items():
            for future, index in futures_list:
                try:
                    response = future.result()
                    results_dict[metric][index] = response
                except Exception as e:
                    results_dict[metric][index] = "Error"
                    error_list.append((metric, index, f"Error: {e}"))

    for metric, responses in results_dict.items():
        results_dict[metric] = [
            _instance_evaluation.handle_response(response, metric)
            for response in responses
        ]
    if error_list:
        _LOGGER.warning(
            f"{len(error_list)} errors encountered during evaluation. Continue to"
            " compute summary metrics for the rest of the dataset."
        )
        for metric_name, index, error in error_list:
            _LOGGER.warning(
                f"Error encountered for metric {metric_name} at dataset index"
                f" {index}: {error}"
            )
    else:
        _LOGGER.info(
            f"All {total_request_count} metric requests are successfully computed."
        )

    instance_df = pd.DataFrame.from_dict(instance_list)
    metrics_table = _parse_metric_results_to_dataframe(instance_df, results_dict)

    # Aggregate the summary metrics.
    summary_metrics = _aggregate_summary_metrics(evaluation_run_config, metrics_table)

    return evaluation_base.EvalResult(
        summary_metrics=summary_metrics, metrics_table=metrics_table
    )


def _get_baseline_model(evaluation_run_config: evaluation_base.EvaluationRunConfig):
    """Gets the baseline model from the pairwise metrics."""
    pairwise_metric_instances = [
        metric
        for metric in evaluation_run_config.metrics
        if isinstance(metric, pairwise_metric.PairwiseMetric)
    ]
    baseline_models = {
        instance.metric_name: instance.baseline_model
        for instance in pairwise_metric_instances
    }
    if len(set(baseline_models.values())) > 1:
        raise ValueError(
            "Not all `PairwiseMetric` instances have the same `baseline_model`. "
            f"Here are the detected baseline models: `{baseline_models}`. "
            "Please separate pairwise metrics with different baseline models "
            "in different `EvalTask` or use the same baseline model for "
            "all pairwise metrics."
        )
    return pairwise_metric_instances[0].baseline_model


def _convert_metric_prompt_template_example(metrics):
    """Converts string metric names to generic model-based metric instances."""
    updated_metrics = []
    for metric in metrics:
        if metric in constants.Metric.POINTWISE_METRIC_PROMPT_TEMPLATE_EXAMPLE_LIST:
            template = metric_prompt_template_examples.MetricPromptTemplateExamples.get_prompt_template(
                metric
            )
            metric = pointwise_metric.PointwiseMetric(
                metric=metric, metric_prompt_template=template
            )
        elif metric in constants.Metric.PAIRWISE_METRIC_PROMPT_TEMPLATE_EXAMPLE_LIST:
            template = metric_prompt_template_examples.MetricPromptTemplateExamples.get_prompt_template(
                metric
            )
            metric = pairwise_metric.PairwiseMetric(
                metric=metric, metric_prompt_template=template
            )
            _LOGGER.info(
                f"Pairwise metric `{metric.metric_name}` loaded from"
                " `MetricPromptTemplateExamples` does not have `baseline_model`"
                " specified and only supports Bring-Your-Own-Response(BYOR)"
                " evaluation. If you would like to run inference on the baseline model,"
                " please instantiate a `PairwiseMetric` and provide the"
                " `baseline_model` parameter."
            )
        updated_metrics.append(metric)
    return updated_metrics


def _get_rubric_metric_with_idx(
    metrics: List[Union[str, metrics_base._Metric]],
) -> Optional[Tuple[rubric_based_metric.RubricBasedMetric, int]]:
    """Gets the rubric metric with its index in the metrics list."""
    for i, metric in enumerate(metrics):
        if isinstance(metric, rubric_based_metric.RubricBasedMetric):
            return metric, i
    return None


def _safe_deepcopy_rubric_generation_config(
    generation_config: metrics_base.RubricGenerationConfig,
) -> metrics_base.RubricGenerationConfig:
    """Safely deepcopies the rubric generation config of RubricBasedMetric."""
    copied_generation_config = None

    if generation_config:
        generation_config_model = generation_config.model
        # Cannot deepcopy instance of GenerativeModel due to its cygrpc.Channel
        generation_config.model = None
        copied_generation_config = copy.deepcopy(generation_config)

        if generation_config_model:
            copied_generation_config.model = generation_config_model

    return copied_generation_config


def evaluate(
    dataset: "pd.DataFrame",
    metrics: List[Union[str, metrics_base._Metric]],
    *,
    model: Optional[_ModelType] = None,
    runnable: Optional[_RunnableType] = None,
    prompt_template: Optional[Union[str, prompt_template_base.PromptTemplate]] = None,
    metric_column_mapping: Dict[str, str],
    evaluation_service_qps: Optional[float] = None,
    retry_timeout: float = 600.0,
    autorater_config: Optional[evaluation_base.AutoraterConfig] = None,
) -> evaluation_base.EvalResult:
    """Runs the evaluation for metrics.

    Args:
      dataset: The dataset to evaluate.
      metrics: The list of metric names, or Metric instances to
        evaluate. Prompt template is required for PairwiseMetric.
      model: The GenerativeModel instance or a custom model function to generate
        responses to evaluate. If not provided, the evaluation is computed with
        the `response` column in the `dataset`.
      runnable: The runnable to generate responses to evaluate. If not provided, the
        evaluation is computed with the `response` and/or `predicted_trajectory`
        column in the `dataset`.
      prompt_template: A `PromptTemplate` or a prompt template string compatible
        with `PromptTemplate` class with variables that can be formatted with
        dataset columns to create assembled prompts. The variables can be
        represented in curly braces `{variable_name}`, and must be included in the
        dataset columns if specified. The variable names cannot contain spaces.
      metric_column_mapping: An optional dictionary column mapping that
        overrides the metric prompt template input variable names with
        mapped the evaluation dataset column names, used during evaluation.
        For example, if the input_variables of the metric prompt template
        are ["context", "reference"], the metric_column_mapping can be
              {
                  "context": "news_context",
                  "reference": "ground_truth",
                  "response": "model_1_response"
              }
            if the dataset has columns "news_context", "ground_truth" and
            "model_1_response".
      evaluation_service_qps: The custom QPS limit for the evaluation service.
      retry_timeout: How long to keep retrying the evaluation requests for the
        whole evaluation dataset, in seconds.
      autorater_config: The autorater config for model based evaluation.

    Returns:
      EvalResult with summary metrics and a metrics table for per-instance
      metrics.

    Raises:
      ValueError: If the metrics list is empty, or the prompt template is not
        provided for PairwiseMetric, or multiple baseline models are specified for
        PairwiseMetric instances, or both model and dataset model response column
        are present, or the sampling count is not in the range [1, 32], or flipping
        is enabled for pairwise metrics but the required columns are not found in
        the dataset.
    """
    _validate_metrics(metrics)
    metrics = _convert_metric_prompt_template_example(metrics)
    copied_metrics = []
    for metric in metrics:
        if isinstance(metric, pairwise_metric.PairwiseMetric):
            copied_metrics.append(
                pairwise_metric.PairwiseMetric(
                    metric=metric.metric_name,
                    metric_prompt_template=metric.metric_prompt_template,
                    baseline_model=metric.baseline_model,
                    system_instruction=metric.system_instruction,
                    custom_output_config=metric.custom_output_config,
                    autorater_config=metric.autorater_config,
                )
            )
        elif isinstance(metric, rubric_based_metric.RubricBasedMetric):
            copied_metrics.append(
                rubric_based_metric.RubricBasedMetric(
                    generation_config=_safe_deepcopy_rubric_generation_config(
                        metric.generation_config,
                    ),
                    critique_metric=copy.deepcopy(metric.critique_metric),
                )
            )
        else:
            copied_metrics.append(copy.deepcopy(metric))

    evaluation_run_config = evaluation_base.EvaluationRunConfig(
        dataset=dataset.copy(deep=True),
        metrics=copied_metrics,
        metric_column_mapping=copy.deepcopy(metric_column_mapping),
        client=utils.create_evaluation_service_client(),
        evaluation_service_qps=(
            evaluation_service_qps
            if evaluation_service_qps
            else constants.QuotaLimit.EVAL_SERVICE_QPS
        ),
        retry_timeout=retry_timeout,
        autorater_config=autorater_config,
    )

    if prompt_template:
        _assemble_prompt_for_dataset(evaluation_run_config, prompt_template)

    if model and runnable:
        raise ValueError(
            "Only one of model and runnable can be provided in a single evaluation."
        )

    if model:
        _run_model_inference(
            model=model,
            evaluation_run_config=evaluation_run_config,
            response_column_name=constants.Dataset.MODEL_RESPONSE_COLUMN,
        )

    if runnable:
        _run_runnable_inference(
            runnable=runnable,
            evaluation_run_config=evaluation_run_config,
            response_column_name=constants.Dataset.MODEL_RESPONSE_COLUMN,
            predicted_trajectory_column_name=constants.Dataset.PREDICTED_TRAJECTORY_COLUMN,
        )
    _validate_dataset(evaluation_run_config)

    rubric_metric_with_idx = _get_rubric_metric_with_idx(evaluation_run_config.metrics)

    if rubric_metric_with_idx:
        rubric_metric, idx = rubric_metric_with_idx
        eval_dataset_with_rubrics = rubric_metric.generate_rubrics(
            evaluation_run_config.dataset
        )
        for column in eval_dataset_with_rubrics.columns:
            if column not in evaluation_run_config.metric_column_mapping:
                evaluation_run_config.metric_column_mapping[column] = column
        evaluation_run_config.dataset = eval_dataset_with_rubrics
        if not rubric_metric.critique_metric.custom_output_config:
            rubric_metric.critique_metric.custom_output_config = (
                custom_output_config.CustomOutputConfig()
            )
        rubric_metric.critique_metric.custom_output_config.return_raw_output = True
        evaluation_run_config.metrics[idx] = rubric_metric.critique_metric

    pairwise_metric_exists = any(
        isinstance(metric, pairwise_metric.PairwiseMetric)
        for metric in evaluation_run_config.metrics
    )
    if pairwise_metric_exists:
        baseline_model = _get_baseline_model(evaluation_run_config)
        if baseline_model:
            _run_model_inference(
                model=baseline_model,
                evaluation_run_config=evaluation_run_config,
                response_column_name=constants.Dataset.BASELINE_MODEL_RESPONSE_COLUMN,
            )

    _validate_autorater_config(evaluation_run_config, pairwise_metric_exists)
    _validate_metric_column_map(evaluation_run_config)
    t1 = time.perf_counter()
    evaluation_result = _compute_metrics(evaluation_run_config)
    t2 = time.perf_counter()
    _LOGGER.info(f"Evaluation Took:{t2 - t1} seconds")

    return evaluation_result



# -*- coding: utf-8 -*-

# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import List, Optional

from google.protobuf import struct_pb2

from google.auth import credentials as auth_credentials

from google.cloud import aiplatform
from google.cloud.aiplatform import base
from google.cloud.aiplatform import models
from google.cloud.aiplatform import pipeline_jobs
from google.cloud.aiplatform import utils


class ModelEvaluation(base.VertexAiResourceNounWithFutureManager):

    client_class = utils.ModelClientWithOverride
    _resource_noun = "evaluations"
    _delete_method = None
    _getter_method = "get_model_evaluation"
    _list_method = "list_model_evaluations"
    _parse_resource_name_method = "parse_model_evaluation_path"
    _format_resource_name_method = "model_evaluation_path"

    @property
    def metrics(self) -> struct_pb2.Value:
        """Gets the evaluation metrics from the Model Evaluation.

        Returns:
            A struct_pb2.Value with model metrics created from the Model Evaluation
        Raises:
            ValueError: If the Model Evaluation doesn't have metrics.
        """
        if self._gca_resource.metrics:
            return self._gca_resource.metrics

        raise ValueError(
            "This ModelEvaluation does not have any metrics, this could be because the Evaluation job failed. Check the logs for details."
        )

    @property
    def _backing_pipeline_job(self) -> Optional["pipeline_jobs.PipelineJob"]:
        """The managed pipeline for this model evaluation job.
        Returns:
            The PipelineJob resource if this evaluation ran from a managed pipeline or None.
        """
        if (
            "metadata" in self._gca_resource
            and "pipeline_job_resource_name" in self._gca_resource.metadata
        ):
            return aiplatform.PipelineJob.get(
                resource_name=self._gca_resource.metadata["pipeline_job_resource_name"],
                credentials=self.credentials,
            )

    def __init__(
        self,
        evaluation_name: str,
        model_id: Optional[str] = None,
        project: Optional[str] = None,
        location: Optional[str] = None,
        credentials: Optional[auth_credentials.Credentials] = None,
    ):
        """Retrieves the ModelEvaluation resource and instantiates its representation.

        Args:
            evaluation_name (str):
                Required. A fully-qualified model evaluation resource name or evaluation ID.
                Example: "projects/123/locations/us-central1/models/456/evaluations/789" or
                "789". If passing only the evaluation ID, model_id must be provided.
            model_id (str):
                Optional. The ID of the model to retrieve this evaluation from. If passing
                only the evaluation ID as evaluation_name, model_id must be provided.
            project (str):
                Optional project to retrieve model evaluation from. If not set, project
                set in aiplatform.init will be used.
            location (str):
                Optional location to retrieve model evaluation from. If not set, location
                set in aiplatform.init will be used.
            credentials: Optional[auth_credentials.Credentials]=None,
                Custom credentials to use to retrieve this model evaluation. If not set,
                credentials set in aiplatform.init will be used.
        """

        super().__init__(
            project=project,
            location=location,
            credentials=credentials,
            resource_name=evaluation_name,
        )

        self._gca_resource = self._get_gca_resource(
            resource_name=evaluation_name,
            parent_resource_name_fields=(
                {models.Model._resource_noun: model_id} if model_id else model_id
            ),
        )

    def delete(self):
        raise NotImplementedError(
            "Deleting a model evaluation has not been implemented yet."
        )

    @classmethod
    def list(
        cls,
        model: str,
        filter: Optional[str] = None,
        order_by: Optional[str] = None,
        enable_simple_view: bool = False,
        project: Optional[str] = None,
        location: Optional[str] = None,
        credentials: Optional[auth_credentials.Credentials] = None,
    ) -> List["ModelEvaluation"]:
        """List all ModelEvaluation resources on the provided model.

        Example Usage:

        aiplatform.ModelEvaluation.list(
            model="projects/123/locations/us-central1/models/456",
        )

        aiplatform.Model.list(
            model="projects/123/locations/us-central1/models/456",
            order_by="create_time desc, display_name"
        )

        Args:
            model (str):
                Required. The resource name of the model to list evaluations for.
                For example: "projects/123/locations/us-central1/models/456".
            filter (str):
                Optional. An expression for filtering the results of the request.
                For field names both snake_case and camelCase are supported.
            order_by (str):
                Optional. A comma-separated list of fields to order by, sorted in
                ascending order. Use "desc" after a field name for descending.
                Supported fields: `display_name`, `create_time`, `update_time`
            project (str):
                Optional. Project to retrieve list from. If not set, project
                set in aiplatform.init will be used.
            location (str):
                Optional. Location to retrieve list from. If not set, location
                set in aiplatform.init will be used.
            credentials (auth_credentials.Credentials):
                Optional. Custom credentials to use to retrieve list. Overrides
                credentials set in aiplatform.init.
            parent (str):
                Optional. The parent resource name if any to retrieve list from.

        Returns:
            List[VertexAiResourceNoun] - A list of SDK resource objects
        """

        return super()._list_with_local_order(
            filter=filter,
            order_by=order_by,
            project=project,
            location=location,
            credentials=credentials,
            parent=model,
        )



# -*- coding: utf-8 -*-
# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

from typing import MutableMapping, MutableSequence

from google.logging.type import log_severity_pb2  # type: ignore
from google.protobuf import duration_pb2  # type: ignore
from google.protobuf import timestamp_pb2  # type: ignore
import proto  # type: ignore

__protobuf__ = proto.module(
    package="google.appengine.logging.v1",
    manifest={
        "LogLine",
        "SourceLocation",
        "SourceReference",
        "RequestLog",
    },
)


class LogLine(proto.Message):
    r"""Application log line emitted while processing a request.

    Attributes:
        time (google.protobuf.timestamp_pb2.Timestamp):
            Approximate time when this log entry was
            made.
        severity (google.logging.type.log_severity_pb2.LogSeverity):
            Severity of this log entry.
        log_message (str):
            App-provided log message.
        source_location (google.cloud.appengine_logging_v1.types.SourceLocation):
            Where in the source code this log message was
            written.
    """

    time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=1,
        message=timestamp_pb2.Timestamp,
    )
    severity: log_severity_pb2.LogSeverity = proto.Field(
        proto.ENUM,
        number=2,
        enum=log_severity_pb2.LogSeverity,
    )
    log_message: str = proto.Field(
        proto.STRING,
        number=3,
    )
    source_location: "SourceLocation" = proto.Field(
        proto.MESSAGE,
        number=4,
        message="SourceLocation",
    )


class SourceLocation(proto.Message):
    r"""Specifies a location in a source code file.

    Attributes:
        file (str):
            Source file name. Depending on the runtime
            environment, this might be a simple name or a
            fully-qualified name.
        line (int):
            Line within the source file.
        function_name (str):
            Human-readable name of the function or method being invoked,
            with optional context such as the class or package name.
            This information is used in contexts such as the logs
            viewer, where a file and line number are less meaningful.
            The format can vary by language. For example:
            ``qual.if.ied.Class.method`` (Java), ``dir/package.func``
            (Go), ``function`` (Python).
    """

    file: str = proto.Field(
        proto.STRING,
        number=1,
    )
    line: int = proto.Field(
        proto.INT64,
        number=2,
    )
    function_name: str = proto.Field(
        proto.STRING,
        number=3,
    )


class SourceReference(proto.Message):
    r"""A reference to a particular snapshot of the source tree used
    to build and deploy an application.

    Attributes:
        repository (str):
            Optional. A URI string identifying the
            repository. Example:
            "https://github.com/GoogleCloudPlatform/kubernetes.git".
        revision_id (str):
            The canonical and persistent identifier of
            the deployed revision. Example (git):
            "0035781c50ec7aa23385dc841529ce8a4b70db1b".
    """

    repository: str = proto.Field(
        proto.STRING,
        number=1,
    )
    revision_id: str = proto.Field(
        proto.STRING,
        number=2,
    )


class RequestLog(proto.Message):
    r"""Complete log information about a single HTTP request to an
    App Engine application.

    Attributes:
        app_id (str):
            Application that handled this request.
        module_id (str):
            Module of the application that handled this
            request.
        version_id (str):
            Version of the application that handled this
            request.
        request_id (str):
            Globally unique identifier for a request,
            which is based on the request start time.
            Request IDs for requests which started later
            will compare greater as strings than those for
            requests which started earlier.
        ip (str):
            Origin IP address.
        start_time (google.protobuf.timestamp_pb2.Timestamp):
            Time when the request started.
        end_time (google.protobuf.timestamp_pb2.Timestamp):
            Time when the request finished.
        latency (google.protobuf.duration_pb2.Duration):
            Latency of the request.
        mega_cycles (int):
            Number of CPU megacycles used to process
            request.
        method (str):
            Request method. Example: ``"GET"``, ``"HEAD"``, ``"PUT"``,
            ``"POST"``, ``"DELETE"``.
        resource (str):
            Contains the path and query portion of the URL that was
            requested. For example, if the URL was
            "http://example.com/app?name=val", the resource would be
            "/app?name=val". The fragment identifier, which is
            identified by the ``#`` character, is not included.
        http_version (str):
            HTTP version of request. Example: ``"HTTP/1.1"``.
        status (int):
            HTTP response status code. Example: 200, 404.
        response_size (int):
            Size in bytes sent back to client by request.
        referrer (str):
            Referrer URL of request.
        user_agent (str):
            User agent that made the request.
        nickname (str):
            The logged-in user who made the request.

            Most likely, this is the part of the user's email before the
            ``@`` sign. The field value is the same for different
            requests from the same user, but different users can have
            similar names. This information is also available to the
            application via the App Engine Users API.

            This field will be populated starting with App Engine
            1.9.21.
        url_map_entry (str):
            File or class that handled the request.
        host (str):
            Internet host and port number of the resource
            being requested.
        cost (float):
            An indication of the relative cost of serving
            this request.
        task_queue_name (str):
            Queue name of the request, in the case of an
            offline request.
        task_name (str):
            Task name of the request, in the case of an
            offline request.
        was_loading_request (bool):
            Whether this was a loading request for the
            instance.
        pending_time (google.protobuf.duration_pb2.Duration):
            Time this request spent in the pending
            request queue.
        instance_index (int):
            If the instance processing this request
            belongs to a manually scaled module, then this
            is the 0-based index of the instance. Otherwise,
            this value is -1.
        finished (bool):
            Whether this request is finished or active.
        first (bool):
            Whether this is the first ``RequestLog`` entry for this
            request. If an active request has several ``RequestLog``
            entries written to Stackdriver Logging, then this field will
            be set for one of them.
        instance_id (str):
            An identifier for the instance that handled
            the request.
        line (MutableSequence[google.cloud.appengine_logging_v1.types.LogLine]):
            A list of log lines emitted by the
            application while serving this request.
        app_engine_release (str):
            App Engine release version.
        trace_id (str):
            Stackdriver Trace identifier for this
            request.
        trace_sampled (bool):
            If true, the value in the 'trace_id' field was sampled for
            storage in a trace backend.
        source_reference (MutableSequence[google.cloud.appengine_logging_v1.types.SourceReference]):
            Source code for the application that handled
            this request. There can be more than one source
            reference per deployed application if source
            code is distributed among multiple repositories.
    """

    app_id: str = proto.Field(
        proto.STRING,
        number=1,
    )
    module_id: str = proto.Field(
        proto.STRING,
        number=37,
    )
    version_id: str = proto.Field(
        proto.STRING,
        number=2,
    )
    request_id: str = proto.Field(
        proto.STRING,
        number=3,
    )
    ip: str = proto.Field(
        proto.STRING,
        number=4,
    )
    start_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=6,
        message=timestamp_pb2.Timestamp,
    )
    end_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=7,
        message=timestamp_pb2.Timestamp,
    )
    latency: duration_pb2.Duration = proto.Field(
        proto.MESSAGE,
        number=8,
        message=duration_pb2.Duration,
    )
    mega_cycles: int = proto.Field(
        proto.INT64,
        number=9,
    )
    method: str = proto.Field(
        proto.STRING,
        number=10,
    )
    resource: str = proto.Field(
        proto.STRING,
        number=11,
    )
    http_version: str = proto.Field(
        proto.STRING,
        number=12,
    )
    status: int = proto.Field(
        proto.INT32,
        number=13,
    )
    response_size: int = proto.Field(
        proto.INT64,
        number=14,
    )
    referrer: str = proto.Field(
        proto.STRING,
        number=15,
    )
    user_agent: str = proto.Field(
        proto.STRING,
        number=16,
    )
    nickname: str = proto.Field(
        proto.STRING,
        number=40,
    )
    url_map_entry: str = proto.Field(
        proto.STRING,
        number=17,
    )
    host: str = proto.Field(
        proto.STRING,
        number=20,
    )
    cost: float = proto.Field(
        proto.DOUBLE,
        number=21,
    )
    task_queue_name: str = proto.Field(
        proto.STRING,
        number=22,
    )
    task_name: str = proto.Field(
        proto.STRING,
        number=23,
    )
    was_loading_request: bool = proto.Field(
        proto.BOOL,
        number=24,
    )
    pending_time: duration_pb2.Duration = proto.Field(
        proto.MESSAGE,
        number=25,
        message=duration_pb2.Duration,
    )
    instance_index: int = proto.Field(
        proto.INT32,
        number=26,
    )
    finished: bool = proto.Field(
        proto.BOOL,
        number=27,
    )
    first: bool = proto.Field(
        proto.BOOL,
        number=42,
    )
    instance_id: str = proto.Field(
        proto.STRING,
        number=28,
    )
    line: MutableSequence["LogLine"] = proto.RepeatedField(
        proto.MESSAGE,
        number=29,
        message="LogLine",
    )
    app_engine_release: str = proto.Field(
        proto.STRING,
        number=38,
    )
    trace_id: str = proto.Field(
        proto.STRING,
        number=39,
    )
    trace_sampled: bool = proto.Field(
        proto.BOOL,
        number=43,
    )
    source_reference: MutableSequence["SourceReference"] = proto.RepeatedField(
        proto.MESSAGE,
        number=41,
        message="SourceReference",
    )


__all__ = tuple(sorted(__protobuf__.manifest))



# -*- coding: utf-8 -*-
# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

from typing import MutableMapping, MutableSequence

import proto  # type: ignore


__protobuf__ = proto.module(
    package="google.cloud.aiplatform.v1",
    manifest={
        "DeploymentStage",
    },
)


class DeploymentStage(proto.Enum):
    r"""Stage field indicating the current progress of a deployment.

    Values:
        DEPLOYMENT_STAGE_UNSPECIFIED (0):
            Default value. This value is unused.
        STARTING_DEPLOYMENT (5):
            The deployment is initializing and setting up
            the environment.
        PREPARING_MODEL (6):
            The deployment is preparing the model assets.
        CREATING_SERVING_CLUSTER (7):
            The deployment is creating the underlying
            serving cluster.
        ADDING_NODES_TO_CLUSTER (8):
            The deployment is adding nodes to the serving
            cluster.
        GETTING_CONTAINER_IMAGE (9):
            The deployment is getting the container image
            for the model server.
        STARTING_MODEL_SERVER (3):
            The deployment is starting the model server.
        FINISHING_UP (4):
            The deployment is performing finalization
            steps.
        DEPLOYMENT_TERMINATED (10):
            The deployment has terminated.
        SUCCESSFULLY_DEPLOYED (11):
            The deployment has succeeded.
        FAILED_TO_DEPLOY (12):
            The deployment has failed.
    """

    DEPLOYMENT_STAGE_UNSPECIFIED = 0
    STARTING_DEPLOYMENT = 5
    PREPARING_MODEL = 6
    CREATING_SERVING_CLUSTER = 7
    ADDING_NODES_TO_CLUSTER = 8
    GETTING_CONTAINER_IMAGE = 9
    STARTING_MODEL_SERVER = 3
    FINISHING_UP = 4
    DEPLOYMENT_TERMINATED = 10
    SUCCESSFULLY_DEPLOYED = 11
    FAILED_TO_DEPLOY = 12


__all__ = tuple(sorted(__protobuf__.manifest))



import os
import requests
from dotenv import load_dotenv
from typing import Dict, Any

load_dotenv()

class AlphaVantageApiClient:
    """API client for Alpha Vantage"""

    def __init__(self):
        self.api_key = os.getenv("ALPHA_VANTAGE_API_KEY")
        self.base_url = "https://www.alphavantage.co/query"

    def get_time_series_daily(self, symbol: str, output_size: str = "compact") -> Dict[str, Any]:
        """
        Get time series daily data for a given stock symbol.
        """
        if not self.api_key:
            return {"error": "Alpha Vantage API key not found."}

        params = {
            "function": "TIME_SERIES_DAILY",
            "symbol": symbol,
            "outputsize": output_size,
            "apikey": self.api_key,
        }
        try:
            response = requests.get(self.base_url, params=params)
            response.raise_for_status()
            return response.json()
        except requests.exceptions.RequestException as e:
            return {"error": f"Error fetching data from Alpha Vantage: {e}"}



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Class for error status."""


class Status:
    """A status, comprising a code and a message.

    See: `Cloud APIs Errors <https://cloud.google.com/apis/design/errors>`_

    This is a thin wrapper for ``google.rpc.status_pb2.Status``.

    :type status_pb: google.rpc.status_pb2.Status
    :param status_pb: The status protocol buffer.
    """

    def __init__(self, status_pb):
        self.status_pb = status_pb

    @property
    def code(self):
        """The status code.

        Values are defined in ``google.rpc.code_pb2.Code``.

        See: `google.rpc.Code
        <https://github.com/googleapis/googleapis/blob/main/google/rpc/code.proto>`_

        :rtype: int
        :returns: The status code.
        """
        return self.status_pb.code

    @property
    def message(self):
        """A human readable status message.

        :rypte: str
        :returns: The status message.
        """
        return self.status_pb.message

    def __repr__(self):
        return repr(self.status_pb)

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return self.status_pb == other.status_pb
        return NotImplemented

    def __ne__(self, other):
        return not self == other



# -*- coding: utf-8 -*-

# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from copy import deepcopy

from google.cloud.aiplatform import base
from google.cloud.aiplatform import initializer as aiplatform_initializer
from google.cloud.aiplatform.compat.services import dataset_service_client
from vertexai.generative_models import (
    Content,
    Image,
    Part,
    GenerativeModel,
    GenerationConfig,
    SafetySetting,
    Tool,
    ToolConfig,
)
from vertexai.generative_models._generative_models import (
    _to_content,
    _validate_generate_content_parameters,
    _reconcile_model_name,
    _get_resource_name_from_model_name,
    ContentsType,
    GenerationConfigType,
    GenerationResponse,
    PartsType,
    SafetySettingsType,
)

import re
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Union,
)

_LOGGER = base.Logger(__name__)

DEFAULT_MODEL_NAME = "gemini-1.5-flash-002"
VARIABLE_NAME_REGEX = r"(\{[^\W0-9]\w*\})"


class Prompt:
    """A prompt which may be a template with variables.

    The `Prompt` class allows users to define a template string with
    variables represented in curly braces `{variable}`. The variable
    name must be a valid Python variable name (no spaces, must start with a
    letter). These placeholders can be replaced with specific values using the
    `assemble_contents` method, providing flexibility in generating dynamic prompts.

    Usage:
        Generate content from a single set of variables:
        ```
        prompt = Prompt(
            prompt_data="Hello, {name}! Today is {day}. How are you?",
            variables=[{"name": "Alice", "day": "Monday"}]
            generation_config=GenerationConfig(
                temperature=0.1,
                top_p=0.95,
                top_k=20,
                candidate_count=1,
                max_output_tokens=100,
            ),
            model_name="gemini-1.0-pro-002",
            safety_settings=[SafetySetting(
                category=SafetySetting.HarmCategory.HARM_CATEGORY_DANGEROUS_CONTENT,
                threshold=SafetySetting.HarmBlockThreshold.BLOCK_MEDIUM_AND_ABOVE,
                method=SafetySetting.HarmBlockMethod.SEVERITY,
            )],
            system_instruction="Please answer in a short sentence.",
        )

        # Generate content using the assembled prompt.
        prompt.generate_content(
            contents=prompt.assemble_contents(**prompt.variables)
        )
        ```
        Generate content with multiple sets of variables:
        ```
        prompt = Prompt(
            prompt_data="Hello, {name}! Today is {day}. How are you?",
            variables=[
                {"name": "Alice", "day": "Monday"},
                {"name": "Bob", "day": "Tuesday"},
            ],
            generation_config=GenerationConfig(
                temperature=0.1,
                top_p=0.95,
                top_k=20,
                candidate_count=1,
                max_output_tokens=100,
            ),
            model_name="gemini-1.0-pro-002",
            safety_settings=[SafetySetting(
                category=SafetySetting.HarmCategory.HARM_CATEGORY_DANGEROUS_CONTENT,
                threshold=SafetySetting.HarmBlockThreshold.BLOCK_MEDIUM_AND_ABOVE,
                method=SafetySetting.HarmBlockMethod.SEVERITY,
            )],
            system_instruction="Please answer in a short sentence.",
        )

        # Generate content using the assembled prompt for each variable set.
        for i in range(len(prompt.variables)):
            prompt.generate_content(
                contents=prompt.assemble_contents(**prompt.variables[i])
            )
        ```
    """

    def __init__(
        self,
        prompt_data: Optional[PartsType] = None,
        *,
        variables: Optional[List[Dict[str, PartsType]]] = None,
        prompt_name: Optional[str] = None,
        generation_config: Optional[GenerationConfig] = None,
        model_name: Optional[str] = None,
        safety_settings: Optional[SafetySetting] = None,
        system_instruction: Optional[PartsType] = None,
        tools: Optional[List[Tool]] = None,
        tool_config: Optional[ToolConfig] = None,
    ):
        """Initializes the Prompt with a given prompt, and variables.

        Args:
            prompt: A PartsType prompt which may be a template with variables or a prompt with no variables.
            variables: A list of dictionaries containing the variable names and values.
            prompt_name: The display name of the prompt, if stored in an online resource.
            generation_config: A GenerationConfig object containing parameters for generation.
            model_name: Model Garden model resource name.
                Alternatively, a tuned model endpoint resource name can be provided.
                If no model is provided, the default latest model will be used.
            safety_settings: A SafetySetting object containing safety settings for generation.
            system_instruction: A PartsType object representing the system instruction.
            tools: A list of Tool objects for function calling.
            tool_config: A ToolConfig object for function calling.
        """
        self._prompt_data = None
        self._variables = None
        self._model_name = None
        self._generation_config = None
        self._safety_settings = None
        self._system_instruction = None
        self._tools = None
        self._tool_config = None

        # Prompt Management
        self._dataset_client_value = None
        self._dataset = None
        self._prompt_name = None
        self._version_id = None
        self._version_name = None

        self.prompt_data = prompt_data
        self.variables = variables if variables else [{}]
        self.prompt_name = prompt_name
        self.model_name = model_name
        self.generation_config = generation_config
        self.safety_settings = safety_settings
        self.system_instruction = system_instruction
        self.tools = tools
        self.tool_config = tool_config

    @property
    def prompt_data(self) -> Optional[PartsType]:
        return self._prompt_data

    @property
    def variables(self) -> Optional[List[Dict[str, PartsType]]]:
        return self._variables

    @property
    def prompt_name(self) -> Optional[str]:
        return self._prompt_name

    @property
    def generation_config(self) -> Optional[GenerationConfig]:
        return self._generation_config

    @property
    def model_name(self) -> Optional[str]:
        if self._model_name:
            return self._model_name
        else:
            return Prompt._format_model_resource_name(DEFAULT_MODEL_NAME)

    @property
    def safety_settings(self) -> Optional[List[SafetySetting]]:
        return self._safety_settings

    @property
    def system_instruction(self) -> Optional[PartsType]:
        return self._system_instruction

    @property
    def tools(self) -> Optional[List[Tool]]:
        return self._tools

    @property
    def tool_config(self) -> Optional[ToolConfig]:
        return self._tool_config

    @property
    def prompt_id(self) -> Optional[str]:
        if self._dataset:
            return self._dataset.name.split("/")[-1]
        return None

    @property
    def version_id(self) -> Optional[str]:
        return self._version_id

    @property
    def version_name(self) -> Optional[str]:
        return self._version_name

    @prompt_data.setter
    def prompt_data(self, prompt_data: Optional[PartsType]) -> None:
        """Overwrites the existing saved local prompt_data.

        Args:
            prompt_data: A PartsType prompt.
        """
        if prompt_data is not None:
            self._validate_parts_type_data(prompt_data)
        self._prompt_data = prompt_data

    @variables.setter
    def variables(self, variables: List[Dict[str, PartsType]]) -> None:
        """Overwrites the existing saved local variables.

        Args:
            variables: A list of dictionaries containing the variable names and values.
        """
        if isinstance(variables, list):
            for i in range(len(variables)):
                variables[i] = variables[i].copy()
                Prompt._format_variable_value_to_parts(variables[i])
            self._variables = variables
        else:
            raise TypeError(
                f"Variables must be a list of dictionaries, not {type(variables)}"
            )

    @prompt_name.setter
    def prompt_name(self, prompt_name: Optional[str]) -> None:
        """Overwrites the existing saved local prompt_name."""
        if prompt_name:
            self._prompt_name = prompt_name
        else:
            self._prompt_name = None

    @model_name.setter
    def model_name(self, model_name: Optional[str]) -> None:
        """Overwrites the existing saved local model_name."""
        if model_name:
            self._model_name = Prompt._format_model_resource_name(model_name)
        else:
            self._model_name = None

    def _format_model_resource_name(model_name: Optional[str]) -> str:
        """Formats the model resource name."""
        project = aiplatform_initializer.global_config.project
        location = aiplatform_initializer.global_config.location
        model_name = _reconcile_model_name(model_name, project, location)

        prediction_resource_name = _get_resource_name_from_model_name(
            model_name, project, location
        )
        return prediction_resource_name

    def _validate_configs(
        self,
        generation_config: Optional[GenerationConfig] = None,
        safety_settings: Optional[SafetySetting] = None,
        system_instruction: Optional[PartsType] = None,
        tools: Optional[List[Tool]] = None,
        tool_config: Optional[ToolConfig] = None,
    ):
        generation_config = generation_config or self._generation_config
        safety_settings = safety_settings or self._safety_settings
        tools = tools or self._tools
        tool_config = tool_config or self._tool_config
        system_instruction = system_instruction or self._system_instruction
        return _validate_generate_content_parameters(
            contents="test",
            generation_config=generation_config,
            safety_settings=safety_settings,
            system_instruction=system_instruction,
            tools=tools,
            tool_config=tool_config,
        )

    @generation_config.setter
    def generation_config(self, generation_config: Optional[GenerationConfig]) -> None:
        """Overwrites the existing saved local generation_config.

        Args:
            generation_config: A GenerationConfig object containing parameters for generation.
        """
        self._validate_configs(generation_config=generation_config)
        self._generation_config = generation_config

    @safety_settings.setter
    def safety_settings(self, safety_settings: Optional[SafetySetting]) -> None:
        """Overwrites the existing saved local safety_settings.

        Args:
            safety_settings: A SafetySetting object containing safety settings for generation.
        """
        self._validate_configs(safety_settings=safety_settings)
        self._safety_settings = safety_settings

    @system_instruction.setter
    def system_instruction(self, system_instruction: Optional[PartsType]) -> None:
        """Overwrites the existing saved local system_instruction.

        Args:
            system_instruction: A PartsType object representing the system instruction.
        """
        if system_instruction:
            self._validate_parts_type_data(system_instruction)
        self._system_instruction = system_instruction

    @tools.setter
    def tools(self, tools: Optional[List[Tool]]) -> None:
        """Overwrites the existing saved local tools.

        Args:
            tools: A list of Tool objects for function calling.
        """
        self._validate_configs(tools=tools)
        self._tools = tools

    @tool_config.setter
    def tool_config(self, tool_config: Optional[ToolConfig] = None) -> None:
        """Overwrites the existing saved local tool_config.

        Args:
            tool_config: A ToolConfig object for function calling.
        """
        self._validate_configs(tool_config=tool_config)
        self._tool_config = tool_config

    def _format_variable_value_to_parts(variables_dict: Dict[str, PartsType]) -> None:
        """Formats the variables values to be List[Part].

        Args:
            variables_dict: A single dictionary containing the variable names and values.

        Raises:
            TypeError: If a variable value is not a PartsType Object.
        """
        for key in variables_dict.keys():
            # Disallow Content as variable value.
            if isinstance(variables_dict[key], Content):
                raise TypeError(
                    "Variable values must be a PartsType object, not Content"
                )

            # Rely on type checks in _to_content for validation.
            content = Content._from_gapic(_to_content(value=variables_dict[key]))
            variables_dict[key] = content.parts

    def _validate_parts_type_data(self, data: Any) -> None:
        """
        Args:
            prompt_data: The prompt input to validate

        Raises:
            TypeError: If prompt_data is not a PartsType Object.
        """
        # Disallow Content as prompt_data.
        if isinstance(data, Content):
            raise TypeError("Prompt data must be a PartsType object, not Content")

        # Rely on type checks in _to_content.
        _to_content(value=data)

    def assemble_contents(self, **variables_dict: PartsType) -> List[Content]:
        """Returns the prompt data, as a List[Content], assembled with variables if applicable.
        Can be ingested into model.generate_content to make API calls.

        Returns:
            A List[Content] prompt.
        Usage:
            ```
            prompt = Prompt(
                prompt_data="Hello, {name}! Today is {day}. How are you?",
            )

            model.generate_content(
                contents=prompt.assemble_contents(name="Alice", day="Monday")
            )
            ```
        """
        # If prompt_data is None, throw an error.
        if self.prompt_data is None:
            raise ValueError("prompt_data must not be empty.")

        variables_dict = variables_dict.copy()

        # If there are no variables, return the prompt_data as a Content object.
        if not variables_dict:
            return [Content._from_gapic(_to_content(value=self.prompt_data))]

        # Step 1) Convert the variables values to List[Part].
        Prompt._format_variable_value_to_parts(variables_dict)

        # Step 2) Assemble the prompt.
        # prompt_data must have been previously validated using _validate_parts_type_data.
        assembled_prompt = []
        assembled_variables_cnt = {}
        if isinstance(self.prompt_data, list):
            # User inputted a List of Parts as prompt_data.
            for part in self.prompt_data:
                assembled_prompt.extend(
                    self._assemble_singular_part(
                        part, variables_dict, assembled_variables_cnt
                    )
                )
        else:
            # User inputted a single str, Image, or Part as prompt_data.
            assembled_prompt.extend(
                self._assemble_singular_part(
                    self.prompt_data, variables_dict, assembled_variables_cnt
                )
            )

        # Step 3) Simplify adjacent string Parts
        simplified_assembled_prompt = [assembled_prompt[0]]
        for i in range(1, len(assembled_prompt)):
            # If the previous Part and current Part is a string, concatenate them.
            try:
                prev_text = simplified_assembled_prompt[-1].text
                curr_text = assembled_prompt[i].text
                if isinstance(prev_text, str) and isinstance(curr_text, str):
                    simplified_assembled_prompt[-1] = Part.from_text(
                        prev_text + curr_text
                    )
                else:
                    simplified_assembled_prompt.append(assembled_prompt[i])
            except AttributeError:
                simplified_assembled_prompt.append(assembled_prompt[i])
                continue

        # Step 4) Validate that all variables were used, if specified.
        for key in variables_dict:
            if key not in assembled_variables_cnt:
                raise ValueError(f"Variable {key} is not present in prompt_data.")

        assemble_cnt_msg = "Assembled prompt replacing: "
        for key in assembled_variables_cnt:
            assemble_cnt_msg += (
                f"{assembled_variables_cnt[key]} instances of variable {key}, "
            )
        if assemble_cnt_msg[-2:] == ", ":
            assemble_cnt_msg = assemble_cnt_msg[:-2]
        _LOGGER.info(assemble_cnt_msg)

        # Step 5) Wrap List[Part] as a single Content object.
        return [
            Content(
                parts=simplified_assembled_prompt,
                role="user",
            )
        ]

    def _assemble_singular_part(
        self,
        prompt_data_part: Union[str, Image, Part],
        formatted_variables_set: Dict[str, List[Part]],
        assembled_variables_cnt: Dict[str, int],
    ) -> List[Part]:
        """Assemble a str, Image, or Part."""
        if isinstance(prompt_data_part, Image):
            # Templating is not supported for Image prompt_data.
            return [Part.from_image(prompt_data_part)]
        elif isinstance(prompt_data_part, str):
            # Assemble a single string
            return self._assemble_single_str(
                prompt_data_part, formatted_variables_set, assembled_variables_cnt
            )
        elif isinstance(prompt_data_part, Part):
            # If the Part is a text Part, assemble it.
            try:
                text = prompt_data_part.text
            except AttributeError:
                return [prompt_data_part]
            return self._assemble_single_str(
                text, formatted_variables_set, assembled_variables_cnt
            )

    def _assemble_single_str(
        self,
        prompt_data_str: str,
        formatted_variables_set: Dict[str, List[Part]],
        assembled_variables_cnt: Dict[str, int],
    ) -> List[Part]:
        """Assemble a single string with 0 or more variables within the string."""
        # Step 1) Find and isolate variables as their own string.
        prompt_data_str_split = re.split(VARIABLE_NAME_REGEX, prompt_data_str)

        assembled_data = []
        # Step 2) Assemble variables with their values, creating a list of Parts.
        for s in prompt_data_str_split:
            if not s:
                continue
            variable_name = s[1:-1]
            if (
                re.match(VARIABLE_NAME_REGEX, s)
                and variable_name in formatted_variables_set
            ):
                assembled_data.extend(formatted_variables_set[variable_name])
                assembled_variables_cnt[variable_name] = (
                    assembled_variables_cnt.get(variable_name, 0) + 1
                )
            else:
                assembled_data.append(Part.from_text(s))

        return assembled_data

    def generate_content(
        self,
        contents: ContentsType,
        *,
        generation_config: Optional[GenerationConfigType] = None,
        safety_settings: Optional[SafetySettingsType] = None,
        model_name: Optional[str] = None,
        tools: Optional[List["Tool"]] = None,
        tool_config: Optional["ToolConfig"] = None,
        stream: bool = False,
        system_instruction: Optional[PartsType] = None,
    ) -> Union[
        "GenerationResponse",
        Iterable["GenerationResponse"],
    ]:
        """Generates content using the saved Prompt configs.

        Args:
            contents: Contents to send to the model.
                Supports either a list of Content objects (passing a multi-turn conversation)
                or a value that can be converted to a single Content object (passing a single message).
                Supports
                * str, Image, Part,
                * List[Union[str, Image, Part]],
                * List[Content]
            generation_config: Parameters for the generation.
            model_name: Prediction model resource name.
            safety_settings: Safety settings as a mapping from HarmCategory to HarmBlockThreshold.
            tools: A list of tools (functions) that the model can try calling.
            tool_config: Config shared for all tools provided in the request.
            stream: Whether to stream the response.
            system_instruction: System instruction to pass to the model.

        Returns:
            A single GenerationResponse object if stream == False
            A stream of GenerationResponse objects if stream == True

        Usage:
            ```
            prompt = Prompt(
                prompt_data="Hello, {name}! Today is {day}. How are you?",
                variables={"name": "Alice", "day": "Monday"},
                generation_config=GenerationConfig(temperature=0.1,),
                system_instruction="Please answer in a short sentence.",
                model_name="gemini-1.0-pro-002",
            )

            prompt.generate_content(
                contents=prompt.assemble_contents(**prompt.variables)
            )
            ```
        """
        if not (model_name or self._model_name):
            _LOGGER.info(
                "No model name specified, falling back to default model: %s",
                self.model_name,
            )
        model_name = model_name or self.model_name

        generation_config = generation_config or self.generation_config
        safety_settings = safety_settings or self.safety_settings
        tools = tools or self.tools
        tool_config = tool_config or self.tool_config
        system_instruction = system_instruction or self.system_instruction

        if not model_name:
            raise ValueError(
                "Model name must be specified to use Prompt.generate_content()"
            )
        model_name = Prompt._format_model_resource_name(model_name)

        model = GenerativeModel(
            model_name=model_name, system_instruction=system_instruction
        )
        return model.generate_content(
            contents=contents,
            generation_config=generation_config,
            safety_settings=safety_settings,
            tools=tools,
            tool_config=tool_config,
            stream=stream,
        )

    @property
    def _dataset_client(self) -> dataset_service_client.DatasetServiceClient:
        if not getattr(self, "_dataset_client_value", None):
            self._dataset_client_value = (
                aiplatform_initializer.global_config.create_client(
                    client_class=dataset_service_client.DatasetServiceClient,
                )
            )
        return self._dataset_client_value

    @classmethod
    def _clone(cls, prompt: "Prompt") -> "Prompt":
        """Returns a copy of the Prompt."""
        return Prompt(
            prompt_data=deepcopy(prompt.prompt_data),
            variables=deepcopy(prompt.variables),
            generation_config=deepcopy(prompt.generation_config),
            safety_settings=deepcopy(prompt.safety_settings),
            tools=deepcopy(prompt.tools),
            tool_config=deepcopy(prompt.tool_config),
            system_instruction=deepcopy(prompt.system_instruction),
            model_name=prompt.model_name,
        )

    def get_unassembled_prompt_data(self) -> PartsType:
        """Returns the prompt data, without any variables replaced."""
        return self.prompt_data

    def __str__(self) -> str:
        """Returns the prompt data as a string, without any variables replaced."""
        return str(self.prompt_data or "")

    def __repr__(self) -> str:
        """Returns a string representation of the unassembled prompt."""
        result = "Prompt("
        if self.prompt_data:
            result += f"prompt_data='{self.prompt_data}', "
        if self.variables and self.variables[0]:
            result += f"variables={self.variables}), "
        if self.system_instruction:
            result += f"system_instruction={self.system_instruction}), "
        if self._model_name:
            # Don't display default model in repr
            result += f"model_name={self._model_name}), "
        if self.generation_config:
            result += f"generation_config={self.generation_config}), "
        if self.safety_settings:
            result += f"safety_settings={self.safety_settings}), "
        if self.tools:
            result += f"tools={self.tools}), "
        if self.tool_config:
            result += f"tool_config={self.tool_config}, "
        if self.prompt_id:
            result += f"prompt_id={self.prompt_id}, "
        if self.version_id:
            result += f"version_id={self.version_id}, "
        if self.prompt_name:
            result += f"prompt_name={self.prompt_name}, "
        if self.version_name:
            result += f"version_name={self.version_name}, "

        # Remove trailing ", "
        if result[-2:] == ", ":
            result = result[:-2]
        result += ")"
        return result



# -*- coding: utf-8 -*-
# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

from typing import MutableMapping, MutableSequence

import proto  # type: ignore

from google.cloud.aiplatform_v1beta1.types import tool
from google.protobuf import struct_pb2  # type: ignore
from google.protobuf import timestamp_pb2  # type: ignore


__protobuf__ = proto.module(
    package="google.cloud.aiplatform.v1beta1",
    manifest={
        "HttpElementLocation",
        "AuthType",
        "Extension",
        "ExtensionManifest",
        "ExtensionOperation",
        "AuthConfig",
        "RuntimeConfig",
        "ExtensionPrivateServiceConnectConfig",
    },
)


class HttpElementLocation(proto.Enum):
    r"""Enum of location an HTTP element can be.

    Values:
        HTTP_IN_UNSPECIFIED (0):
            No description available.
        HTTP_IN_QUERY (1):
            Element is in the HTTP request query.
        HTTP_IN_HEADER (2):
            Element is in the HTTP request header.
        HTTP_IN_PATH (3):
            Element is in the HTTP request path.
        HTTP_IN_BODY (4):
            Element is in the HTTP request body.
        HTTP_IN_COOKIE (5):
            Element is in the HTTP request cookie.
    """

    HTTP_IN_UNSPECIFIED = 0
    HTTP_IN_QUERY = 1
    HTTP_IN_HEADER = 2
    HTTP_IN_PATH = 3
    HTTP_IN_BODY = 4
    HTTP_IN_COOKIE = 5


class AuthType(proto.Enum):
    r"""Type of Auth.

    Values:
        AUTH_TYPE_UNSPECIFIED (0):
            No description available.
        NO_AUTH (1):
            No Auth.
        API_KEY_AUTH (2):
            API Key Auth.
        HTTP_BASIC_AUTH (3):
            HTTP Basic Auth.
        GOOGLE_SERVICE_ACCOUNT_AUTH (4):
            Google Service Account Auth.
        OAUTH (6):
            OAuth auth.
        OIDC_AUTH (8):
            OpenID Connect (OIDC) Auth.
    """

    AUTH_TYPE_UNSPECIFIED = 0
    NO_AUTH = 1
    API_KEY_AUTH = 2
    HTTP_BASIC_AUTH = 3
    GOOGLE_SERVICE_ACCOUNT_AUTH = 4
    OAUTH = 6
    OIDC_AUTH = 8


class Extension(proto.Message):
    r"""Extensions are tools for large language models to access
    external data, run computations, etc.

    Attributes:
        name (str):
            Identifier. The resource name of the
            Extension.
        display_name (str):
            Required. The display name of the Extension.
            The name can be up to 128 characters long and
            can consist of any UTF-8 characters.
        description (str):
            Optional. The description of the Extension.
        create_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. Timestamp when this Extension
            was created.
        update_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. Timestamp when this Extension
            was most recently updated.
        etag (str):
            Optional. Used to perform consistent
            read-modify-write updates. If not set, a blind
            "overwrite" update happens.
        manifest (google.cloud.aiplatform_v1beta1.types.ExtensionManifest):
            Required. Manifest of the Extension.
        extension_operations (MutableSequence[google.cloud.aiplatform_v1beta1.types.ExtensionOperation]):
            Output only. Supported operations.
        runtime_config (google.cloud.aiplatform_v1beta1.types.RuntimeConfig):
            Optional. Runtime config controlling the
            runtime behavior of this Extension.
        tool_use_examples (MutableSequence[google.cloud.aiplatform_v1beta1.types.ToolUseExample]):
            Optional. Examples to illustrate the usage of
            the extension as a tool.
        private_service_connect_config (google.cloud.aiplatform_v1beta1.types.ExtensionPrivateServiceConnectConfig):
            Optional. The PrivateServiceConnect config
            for the extension. If specified, the service
            endpoints associated with the Extension should
            be registered with private network access in the
            provided Service Directory
            (https://cloud.google.com/service-directory/docs/configuring-private-network-access).

            If the service contains more than one endpoint
            with a network, the service will arbitrarilty
            choose one of the endpoints to use for extension
            execution.
    """

    name: str = proto.Field(
        proto.STRING,
        number=1,
    )
    display_name: str = proto.Field(
        proto.STRING,
        number=3,
    )
    description: str = proto.Field(
        proto.STRING,
        number=4,
    )
    create_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=5,
        message=timestamp_pb2.Timestamp,
    )
    update_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=6,
        message=timestamp_pb2.Timestamp,
    )
    etag: str = proto.Field(
        proto.STRING,
        number=7,
    )
    manifest: "ExtensionManifest" = proto.Field(
        proto.MESSAGE,
        number=9,
        message="ExtensionManifest",
    )
    extension_operations: MutableSequence["ExtensionOperation"] = proto.RepeatedField(
        proto.MESSAGE,
        number=11,
        message="ExtensionOperation",
    )
    runtime_config: "RuntimeConfig" = proto.Field(
        proto.MESSAGE,
        number=13,
        message="RuntimeConfig",
    )
    tool_use_examples: MutableSequence[tool.ToolUseExample] = proto.RepeatedField(
        proto.MESSAGE,
        number=15,
        message=tool.ToolUseExample,
    )
    private_service_connect_config: "ExtensionPrivateServiceConnectConfig" = (
        proto.Field(
            proto.MESSAGE,
            number=16,
            message="ExtensionPrivateServiceConnectConfig",
        )
    )


class ExtensionManifest(proto.Message):
    r"""Manifest spec of an Extension needed for runtime execution.

    Attributes:
        name (str):
            Required. Extension name shown to the LLM.
            The name can be up to 128 characters long.
        description (str):
            Required. The natural language description
            shown to the LLM. It should describe the usage
            of the extension, and is essential for the LLM
            to perform reasoning. e.g., if the extension is
            a data store, you can let the LLM know what data
            it contains.
        api_spec (google.cloud.aiplatform_v1beta1.types.ExtensionManifest.ApiSpec):
            Required. Immutable. The API specification
            shown to the LLM.
        auth_config (google.cloud.aiplatform_v1beta1.types.AuthConfig):
            Required. Immutable. Type of auth supported
            by this extension.
    """

    class ApiSpec(proto.Message):
        r"""The API specification shown to the LLM.

        This message has `oneof`_ fields (mutually exclusive fields).
        For each oneof, at most one member field can be set at the same time.
        Setting any member of the oneof automatically clears all other
        members.

        .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

        Attributes:
            open_api_yaml (str):
                The API spec in Open API standard and YAML
                format.

                This field is a member of `oneof`_ ``api_spec``.
            open_api_gcs_uri (str):
                Cloud Storage URI pointing to the OpenAPI
                spec.

                This field is a member of `oneof`_ ``api_spec``.
        """

        open_api_yaml: str = proto.Field(
            proto.STRING,
            number=1,
            oneof="api_spec",
        )
        open_api_gcs_uri: str = proto.Field(
            proto.STRING,
            number=2,
            oneof="api_spec",
        )

    name: str = proto.Field(
        proto.STRING,
        number=1,
    )
    description: str = proto.Field(
        proto.STRING,
        number=2,
    )
    api_spec: ApiSpec = proto.Field(
        proto.MESSAGE,
        number=3,
        message=ApiSpec,
    )
    auth_config: "AuthConfig" = proto.Field(
        proto.MESSAGE,
        number=5,
        message="AuthConfig",
    )


class ExtensionOperation(proto.Message):
    r"""Operation of an extension.

    Attributes:
        operation_id (str):
            Operation ID that uniquely identifies the
            operations among the extension. See: "Operation
            Object" in https://swagger.io/specification/.

            This field is parsed from the OpenAPI spec. For
            HTTP extensions, if it does not exist in the
            spec, we will generate one from the HTTP method
            and path.
        function_declaration (google.cloud.aiplatform_v1beta1.types.FunctionDeclaration):
            Output only. Structured representation of a
            function declaration as defined by the OpenAPI
            Spec.
    """

    operation_id: str = proto.Field(
        proto.STRING,
        number=1,
    )
    function_declaration: tool.FunctionDeclaration = proto.Field(
        proto.MESSAGE,
        number=3,
        message=tool.FunctionDeclaration,
    )


class AuthConfig(proto.Message):
    r"""Auth configuration to run the extension.

    This message has `oneof`_ fields (mutually exclusive fields).
    For each oneof, at most one member field can be set at the same time.
    Setting any member of the oneof automatically clears all other
    members.

    .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

    Attributes:
        api_key_config (google.cloud.aiplatform_v1beta1.types.AuthConfig.ApiKeyConfig):
            Config for API key auth.

            This field is a member of `oneof`_ ``auth_config``.
        http_basic_auth_config (google.cloud.aiplatform_v1beta1.types.AuthConfig.HttpBasicAuthConfig):
            Config for HTTP Basic auth.

            This field is a member of `oneof`_ ``auth_config``.
        google_service_account_config (google.cloud.aiplatform_v1beta1.types.AuthConfig.GoogleServiceAccountConfig):
            Config for Google Service Account auth.

            This field is a member of `oneof`_ ``auth_config``.
        oauth_config (google.cloud.aiplatform_v1beta1.types.AuthConfig.OauthConfig):
            Config for user oauth.

            This field is a member of `oneof`_ ``auth_config``.
        oidc_config (google.cloud.aiplatform_v1beta1.types.AuthConfig.OidcConfig):
            Config for user OIDC auth.

            This field is a member of `oneof`_ ``auth_config``.
        auth_type (google.cloud.aiplatform_v1beta1.types.AuthType):
            Type of auth scheme.
    """

    class ApiKeyConfig(proto.Message):
        r"""Config for authentication with API key.

        Attributes:
            name (str):
                Required. The parameter name of the API key. E.g. If the API
                request is "https://example.com/act?api_key=", "api_key"
                would be the parameter name.
            api_key_secret (str):
                Required. The name of the SecretManager secret version
                resource storing the API key. Format:
                ``projects/{project}/secrets/{secrete}/versions/{version}``

                - If specified, the ``secretmanager.versions.access``
                  permission should be granted to Vertex AI Extension
                  Service Agent
                  (https://cloud.google.com/vertex-ai/docs/general/access-control#service-agents)
                  on the specified resource.
            http_element_location (google.cloud.aiplatform_v1beta1.types.HttpElementLocation):
                Required. The location of the API key.
        """

        name: str = proto.Field(
            proto.STRING,
            number=1,
        )
        api_key_secret: str = proto.Field(
            proto.STRING,
            number=2,
        )
        http_element_location: "HttpElementLocation" = proto.Field(
            proto.ENUM,
            number=3,
            enum="HttpElementLocation",
        )

    class HttpBasicAuthConfig(proto.Message):
        r"""Config for HTTP Basic Authentication.

        Attributes:
            credential_secret (str):
                Required. The name of the SecretManager secret version
                resource storing the base64 encoded credentials. Format:
                ``projects/{project}/secrets/{secrete}/versions/{version}``

                - If specified, the ``secretmanager.versions.access``
                  permission should be granted to Vertex AI Extension
                  Service Agent
                  (https://cloud.google.com/vertex-ai/docs/general/access-control#service-agents)
                  on the specified resource.
        """

        credential_secret: str = proto.Field(
            proto.STRING,
            number=2,
        )

    class GoogleServiceAccountConfig(proto.Message):
        r"""Config for Google Service Account Authentication.

        Attributes:
            service_account (str):
                Optional. The service account that the extension execution
                service runs as.

                - If the service account is specified, the
                  ``iam.serviceAccounts.getAccessToken`` permission should
                  be granted to Vertex AI Extension Service Agent
                  (https://cloud.google.com/vertex-ai/docs/general/access-control#service-agents)
                  on the specified service account.

                - If not specified, the Vertex AI Extension Service Agent
                  will be used to execute the Extension.
        """

        service_account: str = proto.Field(
            proto.STRING,
            number=1,
        )

    class OauthConfig(proto.Message):
        r"""Config for user oauth.

        This message has `oneof`_ fields (mutually exclusive fields).
        For each oneof, at most one member field can be set at the same time.
        Setting any member of the oneof automatically clears all other
        members.

        .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

        Attributes:
            access_token (str):
                Access token for extension endpoint. Only used to propagate
                token from [[ExecuteExtensionRequest.runtime_auth_config]]
                at request time.

                This field is a member of `oneof`_ ``oauth_config``.
            service_account (str):
                The service account used to generate access tokens for
                executing the Extension.

                - If the service account is specified, the
                  ``iam.serviceAccounts.getAccessToken`` permission should
                  be granted to Vertex AI Extension Service Agent
                  (https://cloud.google.com/vertex-ai/docs/general/access-control#service-agents)
                  on the provided service account.

                This field is a member of `oneof`_ ``oauth_config``.
        """

        access_token: str = proto.Field(
            proto.STRING,
            number=1,
            oneof="oauth_config",
        )
        service_account: str = proto.Field(
            proto.STRING,
            number=2,
            oneof="oauth_config",
        )

    class OidcConfig(proto.Message):
        r"""Config for user OIDC auth.

        This message has `oneof`_ fields (mutually exclusive fields).
        For each oneof, at most one member field can be set at the same time.
        Setting any member of the oneof automatically clears all other
        members.

        .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

        Attributes:
            id_token (str):
                OpenID Connect formatted ID token for extension endpoint.
                Only used to propagate token from
                [[ExecuteExtensionRequest.runtime_auth_config]] at request
                time.

                This field is a member of `oneof`_ ``oidc_config``.
            service_account (str):
                The service account used to generate an OpenID Connect
                (OIDC)-compatible JWT token signed by the Google OIDC
                Provider (accounts.google.com) for extension endpoint
                (https://cloud.google.com/iam/docs/create-short-lived-credentials-direct#sa-credentials-oidc).

                - The audience for the token will be set to the URL in the
                  server url defined in the OpenApi spec.

                - If the service account is provided, the service account
                  should grant ``iam.serviceAccounts.getOpenIdToken``
                  permission to Vertex AI Extension Service Agent
                  (https://cloud.google.com/vertex-ai/docs/general/access-control#service-agents).

                This field is a member of `oneof`_ ``oidc_config``.
        """

        id_token: str = proto.Field(
            proto.STRING,
            number=1,
            oneof="oidc_config",
        )
        service_account: str = proto.Field(
            proto.STRING,
            number=2,
            oneof="oidc_config",
        )

    api_key_config: ApiKeyConfig = proto.Field(
        proto.MESSAGE,
        number=2,
        oneof="auth_config",
        message=ApiKeyConfig,
    )
    http_basic_auth_config: HttpBasicAuthConfig = proto.Field(
        proto.MESSAGE,
        number=3,
        oneof="auth_config",
        message=HttpBasicAuthConfig,
    )
    google_service_account_config: GoogleServiceAccountConfig = proto.Field(
        proto.MESSAGE,
        number=4,
        oneof="auth_config",
        message=GoogleServiceAccountConfig,
    )
    oauth_config: OauthConfig = proto.Field(
        proto.MESSAGE,
        number=5,
        oneof="auth_config",
        message=OauthConfig,
    )
    oidc_config: OidcConfig = proto.Field(
        proto.MESSAGE,
        number=7,
        oneof="auth_config",
        message=OidcConfig,
    )
    auth_type: "AuthType" = proto.Field(
        proto.ENUM,
        number=101,
        enum="AuthType",
    )


class RuntimeConfig(proto.Message):
    r"""Runtime configuration to run the extension.

    This message has `oneof`_ fields (mutually exclusive fields).
    For each oneof, at most one member field can be set at the same time.
    Setting any member of the oneof automatically clears all other
    members.

    .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

    Attributes:
        code_interpreter_runtime_config (google.cloud.aiplatform_v1beta1.types.RuntimeConfig.CodeInterpreterRuntimeConfig):
            Code execution runtime configurations for
            code interpreter extension.

            This field is a member of `oneof`_ ``GoogleFirstPartyExtensionConfig``.
        vertex_ai_search_runtime_config (google.cloud.aiplatform_v1beta1.types.RuntimeConfig.VertexAISearchRuntimeConfig):
            Runtime configuration for Vertex AI Search
            extension.

            This field is a member of `oneof`_ ``GoogleFirstPartyExtensionConfig``.
        default_params (google.protobuf.struct_pb2.Struct):
            Optional. Default parameters that will be set for all the
            execution of this extension. If specified, the parameter
            values can be overridden by values in
            [[ExecuteExtensionRequest.operation_params]] at request
            time.

            The struct should be in a form of map with param name as the
            key and actual param value as the value. E.g. If this
            operation requires a param "name" to be set to "abc". you
            can set this to something like {"name": "abc"}.
    """

    class CodeInterpreterRuntimeConfig(proto.Message):
        r"""

        Attributes:
            file_input_gcs_bucket (str):
                Optional. The Cloud Storage bucket for file
                input of this Extension. If specified, support
                input from the Cloud Storage bucket. Vertex
                Extension Custom Code Service Agent should be
                granted file reader to this bucket.
                If not specified, the extension will only accept
                file contents from request body and reject Cloud
                Storage file inputs.
            file_output_gcs_bucket (str):
                Optional. The Cloud Storage bucket for file
                output of this Extension. If specified, write
                all output files to the Cloud Storage bucket.
                Vertex Extension Custom Code Service Agent
                should be granted file writer to this bucket.
                If not specified, the file content will be
                output in response body.
        """

        file_input_gcs_bucket: str = proto.Field(
            proto.STRING,
            number=1,
        )
        file_output_gcs_bucket: str = proto.Field(
            proto.STRING,
            number=2,
        )

    class VertexAISearchRuntimeConfig(proto.Message):
        r"""

        Attributes:
            serving_config_name (str):
                Optional. Vertex AI Search serving config name. Format:
                ``projects/{project}/locations/{location}/collections/{collection}/engines/{engine}/servingConfigs/{serving_config}``
            engine_id (str):
                Optional. Vertex AI Search engine ID. This is used to
                construct the search request. By setting this engine_id, API
                will construct the serving config using the default value to
                call search API for the user. The engine_id and
                serving_config_name cannot both be empty at the same time.
        """

        serving_config_name: str = proto.Field(
            proto.STRING,
            number=1,
        )
        engine_id: str = proto.Field(
            proto.STRING,
            number=2,
        )

    code_interpreter_runtime_config: CodeInterpreterRuntimeConfig = proto.Field(
        proto.MESSAGE,
        number=2,
        oneof="GoogleFirstPartyExtensionConfig",
        message=CodeInterpreterRuntimeConfig,
    )
    vertex_ai_search_runtime_config: VertexAISearchRuntimeConfig = proto.Field(
        proto.MESSAGE,
        number=6,
        oneof="GoogleFirstPartyExtensionConfig",
        message=VertexAISearchRuntimeConfig,
    )
    default_params: struct_pb2.Struct = proto.Field(
        proto.MESSAGE,
        number=4,
        message=struct_pb2.Struct,
    )


class ExtensionPrivateServiceConnectConfig(proto.Message):
    r"""PrivateExtensionConfig configuration for the extension.

    Attributes:
        service_directory (str):
            Required. The Service Directory resource name in which the
            service endpoints associated to the extension are
            registered. Format:
            ``projects/{project_id}/locations/{location_id}/namespaces/{namespace_id}/services/{service_id}``

            - The Vertex AI Extension Service Agent
              (https://cloud.google.com/vertex-ai/docs/general/access-control#service-agents)
              should be granted ``servicedirectory.viewer`` and
              ``servicedirectory.pscAuthorizedService`` roles on the
              resource.
    """

    service_directory: str = proto.Field(
        proto.STRING,
        number=1,
    )


__all__ = tuple(sorted(__protobuf__.manifest))



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Base class for working with Model Garden models."""

import dataclasses
from typing import Dict, Optional, Type, TypeVar
from google.auth import exceptions as auth_exceptions

from google.cloud import aiplatform
from google.cloud.aiplatform import base
from google.cloud.aiplatform import initializer as aiplatform_initializer
from google.cloud.aiplatform import models as aiplatform_models
from google.cloud.aiplatform import _publisher_models
from vertexai._utils import warning_logs

_SUPPORTED_PUBLISHERS = ["google"]

_SHORT_MODEL_ID_TO_TUNING_PIPELINE_MAP = {
    "text-bison": "https://us-kfp.pkg.dev/ml-pipeline/large-language-model-pipelines/tune-large-model/v2.0.0",
    "text-bison-32k": "https://us-kfp.pkg.dev/ml-pipeline/large-language-model-pipelines/tune-large-model/v2.0.0",
    "code-bison": "https://us-kfp.pkg.dev/ml-pipeline/large-language-model-pipelines/tune-large-model/v3.0.0",
    "code-bison-32k": "https://us-kfp.pkg.dev/ml-pipeline/large-language-model-pipelines/tune-large-model/v3.0.0",
    "chat-bison": "https://us-kfp.pkg.dev/ml-pipeline/large-language-model-pipelines/tune-large-chat-model/v3.0.0",
    "chat-bison-32k": "https://us-kfp.pkg.dev/ml-pipeline/large-language-model-pipelines/tune-large-chat-model/v3.0.0",
    "codechat-bison": "https://us-kfp.pkg.dev/ml-pipeline/large-language-model-pipelines/tune-large-chat-model/v3.0.0",
    "codechat-bison-32k": "https://us-kfp.pkg.dev/ml-pipeline/large-language-model-pipelines/tune-large-chat-model/v3.0.0",
    "textembedding-gecko": "https://us-kfp.pkg.dev/ml-pipeline/llm-text-embedding/tune-text-embedding-model/v1.1.3",
    "textembedding-gecko-multilingual": "https://us-kfp.pkg.dev/ml-pipeline/llm-text-embedding/tune-text-embedding-model/v1.1.3",
    "text-embedding-004": "https://us-kfp.pkg.dev/ml-pipeline/llm-text-embedding/tune-text-embedding-model/v1.1.3",
    "text-embedding-005": "https://us-kfp.pkg.dev/ml-pipeline/llm-text-embedding/tune-text-embedding-model/v1.1.4",
    "text-multilingual-embedding-002": "https://us-kfp.pkg.dev/ml-pipeline/llm-text-embedding/tune-text-embedding-model/v1.1.3",
}

_LOGGER = base.Logger(__name__)

T = TypeVar("T", bound="_ModelGardenModel")

# When this module is initialized, _SUBCLASSES contains a mapping of SDK class to the Model Garden instance for that class.
# The key is the SDK class since multiple classes can share a schema URI (i.e. _PreviewTextGenerationModel and TextGenerationModel)
# For example: {"<class 'google.cloud.aiplatform.vertexai.language_models._language_models._TextGenerationModel'>: gs://google-cloud-aiplatform/schema/predict/instance/text_generation_1.0.0.yaml"}
_SUBCLASSES = {}


def _get_model_class_from_schema_uri(
    schema_uri: str,
) -> "_ModelGardenModel":
    """Gets the _ModelGardenModel class for the provided PublisherModel schema uri.

    Args:
        schema_uri (str): The schema_uri for the provided PublisherModel, for example:
        "gs://google-cloud-aiplatform/schema/predict/instance/text_generation_1.0.0.yaml"

    Returns:
        The _ModelGardenModel class associated with the provided schema uri.

    Raises:
        ValueError
            If the provided PublisherModel schema_uri isn't supported by the SDK in Preview.
    """

    for sdk_class in _SUBCLASSES:
        class_schema_uri = _SUBCLASSES[sdk_class]
        if class_schema_uri == schema_uri and "preview" in sdk_class.__module__:
            return sdk_class

    raise ValueError("This model is not supported in Preview by the Vertex SDK.")


@dataclasses.dataclass
class _ModelInfo:
    endpoint_name: str
    interface_class: Type["_ModelGardenModel"]
    publisher_model_resource: _publisher_models._PublisherModel
    tuning_pipeline_uri: Optional[str] = None
    tuning_model_id: Optional[str] = None


def _get_model_info(
    model_id: str,
    schema_to_class_map: Optional[Dict[str, "_ModelGardenModel"]] = None,
    interface_class: Optional[Type["_ModelGardenModel"]] = None,
    publisher_model_res: Optional[_publisher_models._PublisherModel] = None,
    tuned_vertex_model: Optional[aiplatform.Model] = None,
) -> _ModelInfo:
    """Gets the model information by model ID.

    Args:
        model_id (str):
            Identifier of a Model Garden Model. Example: "text-bison@001"
        schema_to_class_map (Dict[str, "_ModelGardenModel"]):
            Mapping of schema URI to model class.

    Returns:
        _ModelInfo:
            Instance of _ModelInfo with details on the provided model_id.

    Raises:
        ValueError:
            If a publisher other than Google is provided in the publisher resource name
            If the provided model doesn't have an associated Publisher Model
            If the model's schema uri is not in the provided schema_to_class_map
    """

    # The default publisher is Google
    if "/" not in model_id:
        model_id = "publishers/google/models/" + model_id

    if not publisher_model_res:
        publisher_model_res = (
            _publisher_models._PublisherModel(  # pylint: disable=protected-access
                resource_name=model_id
            )._gca_resource
        )

    if not publisher_model_res.name.startswith("publishers/google/models/"):
        raise ValueError(
            f"Only Google models are currently supported. {publisher_model_res.name}"
        )
    short_model_id = publisher_model_res.name.rsplit("/", 1)[-1]

    # == "projects/{project}/locations/{location}/publishers/google/models/text-bison@001"
    publisher_model_template = publisher_model_res.publisher_model_template.replace(
        "{user-project}", "{project}"
    )
    if not publisher_model_template:
        raise RuntimeError(
            f"The model does not have an associated Publisher Model. {publisher_model_res.name}"
        )

    if not tuned_vertex_model:
        endpoint_name = publisher_model_template.format(
            project=aiplatform_initializer.global_config.project,
            location=aiplatform_initializer.global_config.location,
        )
    else:
        tuned_model_deployments = tuned_vertex_model.gca_resource.deployed_models
        if len(tuned_model_deployments) == 0:
            # Deploying the model
            endpoint_name = tuned_vertex_model.deploy().resource_name
        else:
            endpoint_name = tuned_model_deployments[0].endpoint

    if short_model_id in _SHORT_MODEL_ID_TO_TUNING_PIPELINE_MAP:
        tuning_pipeline_uri = _SHORT_MODEL_ID_TO_TUNING_PIPELINE_MAP[short_model_id]
        tuning_model_id = publisher_model_template.rsplit("/", 1)[-1]
    else:
        tuning_pipeline_uri = None
        tuning_model_id = None

    if schema_to_class_map:
        interface_class = schema_to_class_map.get(
            publisher_model_res.predict_schemata.instance_schema_uri
        )

        if not interface_class:
            raise ValueError(
                f"Unknown model {publisher_model_res.name}; {schema_to_class_map}"
            )

    return _ModelInfo(
        endpoint_name=endpoint_name,
        interface_class=interface_class,
        publisher_model_resource=publisher_model_res,
        tuning_pipeline_uri=tuning_pipeline_uri,
        tuning_model_id=tuning_model_id,
    )


def _from_pretrained(
    *,
    interface_class: Optional[Type[T]] = None,
    model_name: Optional[str] = None,
    publisher_model: Optional[_publisher_models._PublisherModel] = None,
    tuned_vertex_model: Optional[aiplatform.Model] = None,
) -> T:
    """Loads a _ModelGardenModel.

    Args:
        model_name: Name of the model.

    Returns:
        An instance of a class derieved from `_ModelGardenModel`.

    Raises:
        ValueError: If model_name is unknown.
        ValueError: If model does not support this class.
    """
    if interface_class:
        if not interface_class._INSTANCE_SCHEMA_URI:
            raise ValueError(
                f"Class {interface_class} is not a correct model interface class since it does not have an instance schema URI."
            )

        model_info = _get_model_info(
            model_id=model_name,
            schema_to_class_map={interface_class._INSTANCE_SCHEMA_URI: interface_class},
        )

    else:
        schema_uri = publisher_model._gca_resource.predict_schemata.instance_schema_uri
        interface_class = _get_model_class_from_schema_uri(schema_uri)

        model_info = _get_model_info(
            model_id=model_name,
            interface_class=interface_class,
            publisher_model_res=publisher_model._gca_resource,
            tuned_vertex_model=tuned_vertex_model,
        )

    if not issubclass(model_info.interface_class, interface_class):
        raise ValueError(
            f"{model_name} is of type {model_info.interface_class.__name__} not of type {interface_class.__name__}"
        )

    return model_info.interface_class(
        model_id=model_name,
        endpoint_name=model_info.endpoint_name,
    )


class _ModelGardenModel:
    """Base class for shared methods and properties across Model Garden models."""

    # Subclasses override this attribute to specify their instance schema
    _INSTANCE_SCHEMA_URI: Optional[str] = None

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        _SUBCLASSES[cls] = cls._INSTANCE_SCHEMA_URI

    def __init__(self, model_id: str, endpoint_name: Optional[str] = None):
        """Creates a _ModelGardenModel.

        This constructor should not be called directly.
        Use `{model_class}.from_pretrained(model_name=...)` instead.

        Args:
            model_id: Identifier of a Model Garden Model. Example: "text-bison@001"
            endpoint_name: Vertex Endpoint resource name for the model
        """

        self._model_id = model_id
        self._endpoint_name = endpoint_name
        # TODO(b/280879204)
        # A workaround for not being able to directly instantiate the
        # high-level Endpoint with the PublisherModel resource name.
        self._endpoint = aiplatform.Endpoint._construct_sdk_resource_from_gapic(
            aiplatform_models.gca_endpoint_compat.Endpoint(name=endpoint_name),
        )

    @classmethod
    def from_pretrained(cls: Type[T], model_name: str) -> T:
        """Loads a _ModelGardenModel.

        Args:
            model_name: Name of the model.

        Returns:
            An instance of a class derieved from `_ModelGardenModel`.

        Raises:
            ValueError: If model_name is unknown.
            ValueError: If model does not support this class.
        """
        warning_logs.show_deprecation_warning()
        credential_exception_str = (
            "\nUnable to authenticate your request."
            "\nDepending on your runtime environment, you can complete authentication by:"
            "\n- if in local JupyterLab instance: `!gcloud auth login` "
            "\n- if in Colab:"
            "\n    -`from google.colab import auth`"
            "\n    -`auth.authenticate_user()`"
            "\n- if in service account or other: please follow guidance in https://cloud.google.com/docs/authentication"
        )

        try:
            return _from_pretrained(interface_class=cls, model_name=model_name)
        except auth_exceptions.GoogleAuthError as e:
            raise auth_exceptions.GoogleAuthError(credential_exception_str) from e
